######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains :class:`OutputChannel` class."""
from contextlib import suppress
from dataclasses import InitVar, dataclass
from spine_engine.utils.serialization import deserialize_path, serialize_path


@dataclass
class OutputChannel:
    """Input resource specific export settings."""

    in_label: str
    """Label of input resource."""
    item_name: InitVar[str]
    """Exporter's name."""
    out_label: str = None
    """Label of output resource. Output file name in case of single file export."""
    out_url: dict = None
    """Optional URL for fixed output database."""

    def __post_init__(self, item_name):
        if self.out_label is None:
            _, separator, in_name = self.in_label.partition("@")
            if separator:
                self.out_label = f"{in_name}_exported@{item_name}"
            else:
                self.out_label = f"{self.in_label}_exported@{item_name}"

    def to_dict(self, project_dir):
        """
        Serializes :class:`OutputChannel` into a dictionary.

        Args:
            project_dir (str): project directory

        Returns:
            dict: serialized :class:`OutputChannel`
        """
        channel_dict = {"in_label": self.in_label, "out_label": self.out_label}
        if self.out_url is not None:
            url = self.out_url.copy()
            with suppress(KeyError):
                del url["username"]
            with suppress(KeyError):
                del url["password"]
            if url["dialect"] == "sqlite" and url["database"]:
                url["database"] = serialize_path(url["database"], project_dir)
            channel_dict["out_url"] = url
        return channel_dict

    @staticmethod
    def from_dict(channel_dict, item_name, project_dir):
        """
        Deserializes :class:`OutputChannel` from a dictionary.

        Args:
            channel_dict (dict): serialized :class:`OutputChannel`
            item_name (str): Exporter's name
            project_dir (str): project directory

        Returns:
            OutputChannel: deserialized instance
        """
        channel = OutputChannel(
            channel_dict["in_label"], item_name, channel_dict["out_label"], channel_dict.get("out_url")
        )
        if channel.out_url and not isinstance(channel.out_url["database"], str):
            channel.out_url["database"] = deserialize_path(channel.out_url["database"], project_dir)
        return channel



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Module for view class."""
import os
from PySide6.QtCore import QObject, Qt, QTimer, Signal, Slot
from PySide6.QtGui import QIcon, QPixmap, QStandardItem, QStandardItemModel
from PySide6.QtWidgets import QDialog, QDialogButtonBox, QLabel, QLineEdit, QTextBrowser, QVBoxLayout
from sqlalchemy.engine.url import URL, make_url
from spinetoolbox.fetch_parent import FetchParent
from spinetoolbox.helpers import busy_effect
from spinetoolbox.plotting import PlottingError, plot_db_mngr_items
from spinetoolbox.project_item.project_item import ProjectItem
from spinetoolbox.spine_db_editor.widgets.multi_spine_db_editor import open_db_editor
from spinetoolbox.spine_db_editor.widgets.spine_db_editor import SpineDBEditor
from spinetoolbox.widgets.notification import Notification
from .commands import PinOrUnpinDBValuesCommand
from .executable_item import ExecutableItem
from .item_info import ItemInfo


class View(ProjectItem):
    """View's main class."""

    def __init__(self, name, description, x, y, toolbox, project, pinned_values=None):
        """
        Args:
            name (str): Object name
            description (str): Object description
            x (float): Initial X coordinate of item icon
            y (float): Initial Y coordinate of item icon
            project (SpineToolboxProject): the project this item belongs to
            pinned_values (dict): pinned values by name
        """
        super().__init__(name, description, x, y, project)
        self._toolbox = toolbox
        self._references = {}
        self._pinned_values = pinned_values if pinned_values is not None else {}
        self.pinned_value_model = QStandardItemModel()
        self.reference_model = QStandardItemModel()
        self._spine_ref_icon = QIcon(QPixmap(":/icons/Spine_db_ref_icon.png"))
        self._data_fetchers = []
        self._fetched_parameter_values = []
        self._pin_to_plot = None
        self.populate_pinned_values_list()

    @staticmethod
    def item_type():
        """See base class."""
        return ItemInfo.item_type()

    @property
    def pinned_values(self):
        return self._pinned_values

    def make_signal_handler_dict(self):
        """Returns a dictionary of all shared signals and their handlers.
        This is to enable simpler connecting and disconnecting."""
        s = super().make_signal_handler_dict()
        s[self._properties_ui.pushButton_pin_values.clicked] = self.pin_values
        s[self._properties_ui.pushButton_open_editor.clicked] = self.open_editor
        s[self._properties_ui.pushButton_plot_pinned.clicked] = self.plot_selected_pinned_values
        return s

    def restore_selections(self):
        """Restore selections into shared widgets when this project item is selected."""
        self._properties_ui.treeView_references.setModel(self.reference_model)
        self._properties_ui.treeView_pinned_values.setModel(self.pinned_value_model)
        self._update_buttons_enabled()

    def save_selections(self):
        """Save selections in shared widgets for this project item into instance variables."""
        self._properties_ui.treeView_references.setModel(None)
        self._properties_ui.treeView_pinned_values.setModel(None)

    @Slot(bool)
    def open_editor(self, checked=False):
        """Opens selected db in the Spine database editor."""
        indexes = self._selected_reference_indexes()
        db_urls = self._db_urls(indexes)
        if not db_urls:
            return
        open_db_editor(db_urls, self._toolbox.db_mngr, reuse_existing_editor=True)

    @Slot(bool)
    def pin_values(self, checked=False):
        """Opens selected db in the Spine database editor to pin values."""
        indexes = self._selected_reference_indexes()
        db_urls = self._db_urls(indexes)
        if not db_urls:
            return
        db_editor = SpineDBEditor(self._toolbox.db_mngr, db_urls)
        dialog = _PinValuesDialog(self, db_editor)
        db_editor.pinned_values_updated.connect(dialog.update_pinned_values)
        dialog.data_committed.connect(self._pin_db_values)
        db_editor.show()
        dialog.show()

    def selected_references(self):
        return self._properties_ui.treeView_references.selectedIndexes()

    def selected_pinned_values(self):
        return self._properties_ui.treeView_pinned_values.selectedIndexes()

    def reference_resource_label_from_url(self, url):
        for label, ref_url in self._references.items():
            if str(ref_url) == str(url):
                return label
        self._logger.msg_error.emit(f"<b>{self.name}</b>: Can't find any resource having url {url}.")

    @Slot(str, list)
    def _pin_db_values(self, name, values):
        self._toolbox.undo_stack.push(
            PinOrUnpinDBValuesCommand(self.name, {name: values}, {name: self._pinned_values.get(name)}, self._project)
        )
        self._logger.msg.emit(f"<b>{self.name}</b>: Successfully added pin '{name}'")

    def unpin_selected_pinned_values(self):
        names = [index.data() for index in self._properties_ui.treeView_pinned_values.selectedIndexes()]
        self._toolbox.undo_stack.push(
            PinOrUnpinDBValuesCommand(
                self.name,
                {name: None for name in names},
                {name: self._pinned_values.get(name) for name in names},
                self._project,
            )
        )

    def renamed_selected_pinned_value(self):
        index = self._properties_ui.treeView_pinned_values.selectedIndexes()[0]
        old_name = index.data()
        new_name, ok = _RenamePinDialog.get_new_name(self, old_name, self._toolbox)
        if not ok:
            return
        values = self._pinned_values.get(old_name)
        self._toolbox.undo_stack.push(
            PinOrUnpinDBValuesCommand(
                self.name, {old_name: None, new_name: values}, {old_name: values, new_name: None}, self._project
            )
        )

    def do_pin_db_values(self, values_by_name):
        for name, values in values_by_name.items():
            self._pinned_values[name] = values
            self.populate_pinned_values_list()

    @Slot(bool)
    def plot_selected_pinned_values(self, _checked=False):
        self._properties_ui.pushButton_plot_pinned.setEnabled(False)
        for index in self._selected_pinned_value_indexes():
            self._plot_pinned_value(index)

    @busy_effect
    def _make_plot_widget(self, index):
        self._pin_to_plot = index.data()
        pinned_values = self._pinned_values[index.data()]
        pks_by_resource_label = {}
        for value in pinned_values:
            resource_label, pk = value
            pks_by_resource_label.setdefault(resource_label, []).append(pk)
        fetch_id_base = 0
        for resource_label, pks in pks_by_resource_label.items():
            url_provider_tuple = self._references.get(resource_label)
            if url_provider_tuple is None:
                self._logger.msg_error.emit(f"<b>{self.name}</b>: Can't find any resource with label {resource_label}")
                continue
            url = url_provider_tuple[0]
            db_map = self._toolbox.db_mngr.get_db_map(url, self._toolbox)
            self._data_fetchers += [
                _DatabaseFetch(fetch_id_base + i, self, self._toolbox.db_mngr, db_map, **pk) for i, pk in enumerate(pks)
            ]
            fetch_id_base += len(pks)
        for fetcher in self._data_fetchers:
            fetcher.start()

    def add_to_plot(self, fetch_id, db_map, parameter_record):
        """Adds parameter value to plot widget.

        Args:
            fetch_id (Any): id given to database fetcher
            db_map (DatabaseMapping): database map
            parameter_record (dict): parameter value's database data
        """
        if not self._data_fetchers:
            return
        if not self._fetched_parameter_values:
            self._fetched_parameter_values = len(self._data_fetchers) * [None]
        self._fetched_parameter_values[fetch_id] = (db_map, parameter_record)
        self._plot_if_all_fetched()

    def failed_to_plot(self, fetch_id, db_map, conditions):
        self._logger.msg_warning.emit(
            f"<b>{self.name}</b>: "
            f"Couldn't find any values having {_pk_to_ul(conditions)} in <b>{self._toolbox.db_mngr.name_registry.display_name(db_map.sa_url)}</b>"
        )
        if not self._fetched_parameter_values:
            self._fetched_parameter_values = len(self._data_fetchers) * [None]
        self._fetched_parameter_values[fetch_id] = "skip"
        self._plot_if_all_fetched()

    def _plot_if_all_fetched(self):
        """Plots available data if all fetching has been finished."""
        if any(v is None for v in self._fetched_parameter_values):
            return
        existing_fetched_parameter_values = [v for v in self._fetched_parameter_values if v != "skip"]
        if any(v != "skip" for v in self._fetched_parameter_values):
            plot_db_maps, plot_items = zip(*existing_fetched_parameter_values)
            try:
                plot_widget = plot_db_mngr_items(plot_items, plot_db_maps, self._toolbox.db_mngr.name_registry)
            except PlottingError as error:
                self._logger.msg_error.emit(str(error))
                return
            plot_widget.use_as_window(self._toolbox, self.name + " / " + self._pin_to_plot)
            plot_widget.show()
        else:
            self._logger.msg_warning.emit("Nothing to plot.")
        self._finalize_fetching()

    def _finalize_fetching(self):
        self._fetched_parameter_values.clear()
        for fetcher in self._data_fetchers:
            fetcher.set_obsolete(True)
            fetcher.deleteLater()
        self._data_fetchers.clear()
        self._properties_ui.pushButton_plot_pinned.setEnabled(True)

    def copy_selected_pinned_value_plot_data(self):
        index = self._properties_ui.treeView_pinned_values.selectionModel().currentIndex()
        plot_widget = self._make_plot_widget(index)
        plot_widget.copy_plot_data()

    def _plot_pinned_value(self, index):
        self._make_plot_widget(index)

    def populate_pinned_values_list(self):
        """Populates pinned values list."""
        self.pinned_value_model.clear()
        self.pinned_value_model.setHorizontalHeaderItem(0, QStandardItem("Pinned values"))  # Add header
        for key, values in self._pinned_values.items():
            if not values:
                continue
            qitem = QStandardItem(key)
            qitem.setFlags(~Qt.ItemIsEditable)
            tool_tip = _format_pinned_values(values)
            qitem.setData(tool_tip, Qt.ItemDataRole.ToolTipRole)
            self.pinned_value_model.appendRow(qitem)

    def _update_buttons_enabled(self):
        self._properties_ui.pushButton_open_editor.setEnabled(bool(self._references))
        self._properties_ui.pushButton_pin_values.setEnabled(bool(self._references))
        self._properties_ui.pushButton_plot_pinned.setEnabled(bool(self._references))

    def populate_reference_list(self):
        """Populates reference list."""
        self._update_buttons_enabled()
        self.reference_model.clear()
        self.reference_model.setHorizontalHeaderItem(0, QStandardItem("Available resources"))  # Add header
        for db in sorted(self._references, reverse=True):
            qitem = QStandardItem(db)
            qitem.setFlags(~Qt.ItemIsEditable)
            qitem.setData(self._spine_ref_icon, Qt.ItemDataRole.DecorationRole)
            self.reference_model.appendRow(qitem)

    def upstream_resources_updated(self, resources):
        """See base class."""
        self._references.clear()
        for resource in resources:
            if resource.type_ == "database":
                url = make_url(resource.url)
                self._references[resource.label] = url
            elif resource.type_ == "file":
                filepath = resource.path
                if os.path.splitext(filepath)[1] == ".sqlite":
                    url = URL.create("sqlite", database=filepath)
                    self._references[resource.label] = url
        self.populate_reference_list()

    def replace_resources_from_upstream(self, old, new):
        """See base class."""
        for old_resource, new_resource in zip(old, new):
            if old_resource.type_ == "database" and new_resource.type_ == "database":
                old_label = old_resource.label
                new_label = new_resource.label
            elif old_resource.type_ == "file":
                if (
                    os.path.splitext(old_resource.path)[1] == ".sqlite"
                    and os.path.splitext(new_resource.path)[1] == ".sqlite"
                ):
                    old_label = old_resource.label
                    new_label = new_resource.label
                else:
                    continue
            else:
                continue
            self._references[new_label] = self._references.pop(old_label)
        self.populate_reference_list()

    def _selected_reference_indexes(self):
        """Returns selected references indexes."""
        selection_model = self._properties_ui.treeView_references.selectionModel()
        if not selection_model.hasSelection():
            self._properties_ui.treeView_references.selectAll()
        return self._properties_ui.treeView_references.selectionModel().selectedRows()

    def _selected_pinned_value_indexes(self):
        """Returns selected references indexes."""
        selection_model = self._properties_ui.treeView_pinned_values.selectionModel()
        if not selection_model.hasSelection():
            self._properties_ui.treeView_pinned_values.selectAll()
        return self._properties_ui.treeView_pinned_values.selectionModel().selectedRows()

    def _db_urls(self, indexes):
        """Returns a list of database URLs for given indexes in the reference model."""
        return [self._references[index.data(Qt.ItemDataRole.DisplayRole)] for index in indexes]

    def notify_destination(self, source_item):
        """See base class."""
        if source_item.item_type() == "Tool":
            self._logger.msg.emit(
                "Link established. You can visualize the output from Tool "
                f"<b>{source_item.name}</b> in View <b>{self.name}</b>."
            )
        elif source_item.item_type() == "Data Store":
            self._logger.msg.emit(
                "Link established. You can visualize Data Store "
                f"<b>{source_item.name}</b> in View <b>{self.name}</b>."
            )
        else:
            super().notify_destination(source_item)

    def item_dict(self):
        d = super().item_dict()
        d["pinned_values"] = {name: values for name, values in self._pinned_values.items() if values is not None}
        return d

    @staticmethod
    def from_dict(name, item_dict, toolbox, project):
        description, x, y = ProjectItem.parse_item_dict(item_dict)
        pinned_values = item_dict.get("pinned_values", {})
        for values in pinned_values.values():
            for value in values:
                pks = value[1]
                for pk_key, pk in pks.items():
                    if isinstance(pk, list):
                        pks[pk_key] = tuple(pk)
        return View(name, description, x, y, toolbox, project, pinned_values=pinned_values)


def _pk_to_ul(pk):
    return "".join(["<ul>"] + [f"<li><b>{k}</b>: {v}</li>" for k, v in pk.items() if v] + ["</ul>"])


def _format_pinned_values(values):
    head = """
        <head>
        <style>
            table, th, td {
              border: 1px solid black;
              border-collapse: collapse;
              padding: 5px;
            }
        </style>
        </head>
    """
    tables = []
    for label, pk in values:
        header = [*pk, "source"]
        data = [*pk.values(), label]
        table = "<p><table>"
        table += "<tr>" + "".join([f"<th>{h}</th>" for h in header]) + "</tr>"
        table += "<tr>" + "".join([f"<td>{d}</td>" for d in data]) + "</tr>"
        table += "</table></p>"
        tables.append(table)
    table = "".join(tables)
    return f"<html>{head}<body>{table}</body></html>"


class _PinDialogMixin:
    @property
    def pin_name(self):
        return self._line_edit.text()

    def _check_name_valid(self):
        if self.pin_name in self._view.pinned_values:
            Notification(
                self,
                f"A pin called {self.pin_name} already exists, please select a different name.",
                corner=Qt.BottomRightCorner,
            ).show()
            return False
        return True


class _PinValuesDialog(_PinDialogMixin, QDialog):
    data_committed = Signal(str, list)

    def __init__(self, view, db_editor):
        super().__init__(parent=db_editor)
        self.setWindowTitle("Pin values")
        self._view = view
        self._db_editor = db_editor
        self._pinned_values = []
        outer_layout = QVBoxLayout(self)
        button_box = QDialogButtonBox(self)
        button_box.setStandardButtons(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        self._ok_button = button_box.button(QDialogButtonBox.StandardButton.Ok)
        self._ok_button.setEnabled(False)
        self._line_edit = QLineEdit()
        self._line_edit.setPlaceholderText("Type a name for the pin here...")
        self._text_edit = QTextBrowser()
        self._text_edit.setPlaceholderText(
            "Select parameter values that you want to pin in the Spine DB Editor underneath "
            "(they will be shown here)..."
        )
        outer_layout.addWidget(self._line_edit)
        outer_layout.addWidget(self._text_edit)
        outer_layout.addWidget(button_box)
        button_box.rejected.connect(self.close)
        button_box.rejected.connect(db_editor.close)
        button_box.accepted.connect(self.accept)
        self._line_edit.textEdited.connect(self._update_ok_button_enabled)
        self.setAttribute(Qt.WA_DeleteOnClose)

    @Slot(str)
    def _update_ok_button_enabled(self, _text=None):
        self._ok_button.setEnabled(bool(self._pinned_values) and bool(self.pin_name))

    @Slot(list)
    def update_pinned_values(self, values):
        self._pinned_values = [(self._view.reference_resource_label_from_url(url), pk) for url, pk in values]
        html = _format_pinned_values(self._pinned_values)
        self._text_edit.setHtml(html)
        label = QLabel(html)
        width = label.sizeHint().width()
        doc_margin = self._text_edit.document().documentMargin()
        width += 2 * doc_margin
        self._text_edit.setMinimumWidth(width)
        self._update_ok_button_enabled()

    def accept(self):
        if not self._check_name_valid():
            return
        super().accept()
        self._db_editor.close()
        self.data_committed.emit(self.pin_name, self._pinned_values)


class _RenamePinDialog(_PinDialogMixin, QDialog):
    def __init__(self, view, old_name, parent):
        super().__init__(parent=parent)
        self._view = view
        self.new_name = ""
        self.setWindowTitle("Rename pin")
        outer_layout = QVBoxLayout(self)
        button_box = QDialogButtonBox(self)
        button_box.setStandardButtons(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        self._ok_button = button_box.button(QDialogButtonBox.StandardButton.Ok)
        self._line_edit = QLineEdit()
        self._line_edit.setText(old_name)
        outer_layout.addWidget(QLabel("New name:"))
        outer_layout.addWidget(self._line_edit)
        outer_layout.addWidget(button_box)
        button_box.rejected.connect(self.close)
        button_box.accepted.connect(self.accept)
        self._line_edit.textEdited.connect(self._update_ok_button_enabled)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self._line_edit.setFocus()

    @Slot(str)
    def _update_ok_button_enabled(self, _text=None):
        self._ok_button.setEnabled(bool(self.pin_name))

    def accept(self):
        if not self._check_name_valid():
            return
        super().accept()
        self.new_name = self.pin_name

    @classmethod
    def get_new_name(cls, view, old_name, parent):
        dialog = cls(view, old_name, parent)
        result = dialog.exec()
        return dialog.new_name, result == QDialog.DialogCode.Accepted


class _DatabaseFetch(FetchParent):
    """A data collector that fetches parameter values from database manager."""

    class _Signals(QObject):
        fetched = Signal()

        def __init__(self, database_fetch):
            super().__init__()
            self._database_fetch = database_fetch
            self.fetched.connect(self._finish_fetch)

        @Slot()
        def _finish_fetch(self):
            self._database_fetch.finished()

    def __init__(self, id_, view, db_mngr, db_map, **conditions):
        """
        Args:
            id_ (Any): fetch id
            view (View): View item
            db_mngr (SpineDBManager): database manager
            db_map (DiffDatabaseMapping): database map
            **conditions: filter conditions
        """
        super().__init__()
        self._id = id_
        self._view = view
        self._db_mngr = db_mngr
        self._db_map = db_map
        self._conditions = conditions
        self._value_fetched = False
        self._signals = None

    def start(self):
        """Starts fetching the parameter value."""
        cached_items = self._db_mngr.get_items(self._db_map, self.fetch_item_type)
        cached_match = self._find_one_or_none(cached_items)
        if cached_match is not None:
            self._value_fetched = True
            self._view.add_to_plot(self._id, self._db_map, cached_match)
        else:
            self._signals = _DatabaseFetch._Signals(self)
            self._try_fetching()

    def _try_fetching(self):
        """Repeatedly tries to fetch more data until there is nothing to fetch."""
        if self._db_mngr.can_fetch_more(self._db_map, self):
            self._db_mngr.fetch_more(self._db_map, self)
            QTimer.singleShot(50, self._try_fetching)

    @property
    def fetch_item_type(self):
        """See base class."""
        return "parameter_value"

    def accepts_item(self, item, db_map):
        """See base class."""
        return all(item[key] == value for key, value in self._conditions.items())

    def set_fetched(self, fetched):
        """See base class."""
        super().set_fetched(fetched)
        if fetched and self._signals is not None:
            self._signals.fetched.emit()

    def handle_items_added(self, db_map_data):
        """A callback that stores relevant parameter values.

        Args:
            db_map_data (dict): parameter value data
        """
        data = db_map_data.get(self._db_map)
        if data is None:
            return
        match = self._find_one_or_none(data)
        if match is not None:
            self._value_fetched = True
            self._view.add_to_plot(self._id, self._db_map, match)

    def _find_one_or_none(self, parameter_value_items):
        """Searches parameter value from given items and returns the match.

        Args:
            parameter_value_items (list of dict): parameter value items

        Returns:
            dict: match or None if not found
        """
        for parameter_value in parameter_value_items:
            match = True
            for key, filter_value in self._conditions.items():
                if parameter_value[key] != filter_value:
                    match = False
                    break
            if match:
                return parameter_value
        return None

    def finished(self):
        """Reports plotting failures to View."""
        if not self._value_fetched:
            self._view.failed_to_plot(self._id, self._db_map, self._conditions)
        self._signals.deleteLater()
        self._signals = None



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Undo/redo commands for the DataStore project item."""
from enum import IntEnum, unique
from spine_items.commands import SpineToolboxCommand


@unique
class CommandId(IntEnum):
    UPDATE_URL = 1


class UpdateDSURLCommand(SpineToolboxCommand):
    """Command to update DS url."""

    def __init__(self, ds_name, was_url_valid, project, **kwargs):
        """
        Args:
            ds_name (str): DS name
            was_url_valid (bool): True if previous URL was valid, False otherwise
            project (SpineToolboxProject): project
            **kwargs: url keys and their values
        """
        super().__init__()
        self._ds_name = ds_name
        self._undo_url_is_valid = was_url_valid
        self._redo_kwargs = kwargs
        ds = project.get_item(ds_name)
        self._undo_kwargs = {k: ds.url()[k] for k in kwargs}
        self._project = project
        if len(kwargs) == 1:
            self.setText(f"change {list(kwargs.keys())[0]} of {ds_name}")
        else:
            self.setText(f"change url of {ds_name}")

    def id(self):
        return CommandId.UPDATE_URL.value

    def mergeWith(self, command):
        if (
            not isinstance(command, UpdateDSURLCommand)
            or self._ds_name != command._ds_name
            or command._undo_url_is_valid
            or set(self._redo_kwargs.keys()).isdisjoint(command._redo_kwargs.keys())
        ):
            return False
        diff_key = None
        for key, value in self._redo_kwargs.items():
            old_value = self._undo_kwargs[key]
            if value != old_value:
                if diff_key is not None:
                    return False
                diff_key = key
                break
        else:
            raise RuntimeError("Logic error: nothing changes between undo and redo URLs.")
        if diff_key == "dialect":
            return False
        changed_value = command._redo_kwargs[diff_key]
        if self._redo_kwargs[diff_key] == changed_value:
            return False
        self._redo_kwargs[diff_key] = changed_value
        if self._redo_kwargs[diff_key] == self._undo_kwargs[diff_key]:
            self.setObsolete(True)
        return True

    def redo(self):
        ds = self._project.get_item(self._ds_name)
        ds.do_update_url(**self._redo_kwargs)

    def undo(self):
        ds = self._project.get_item(self._ds_name)
        ds.do_update_url(**self._undo_kwargs)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""MergerPropertiesWidget widget."""
from spinetoolbox.widgets.properties_widget import PropertiesWidgetBase


class MergerPropertiesWidget(PropertiesWidgetBase):
    """Widget for the Data Store Item Properties."""

    def __init__(self, toolbox):
        """
        Args:
            toolbox (ToolboxUI): The toolbox instance where this widget should be embedded
        """
        from ..ui.merger_properties import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(toolbox)
        self.ui = Ui_Form()
        self.ui.setupUi(self)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for DataStore class."""
import gc
import logging
import os
from pathlib import Path
import sys
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from PySide6.QtWidgets import QApplication
from spine_engine.project_item.project_item_resource import database_resource
from spine_items.data_store.data_store import DataStore
from spine_items.data_store.data_store_factory import DataStoreFactory
from spine_items.data_store.item_info import ItemInfo
import spine_items.resources_icons_rc  # pylint: disable=unused-import
from spine_items.utils import convert_to_sqlalchemy_url, database_label
from spinedb_api import create_new_spine_database
from spinetoolbox.helpers import signal_waiter
from tests.mock_helpers import (
    clean_up_toolbox,
    create_mock_project,
    create_mock_toolbox,
    create_toolboxui_with_project,
    mock_finish_project_item_construction,
)


class TestDataStore(unittest.TestCase):
    def test_item_type(self):
        """Tests that the item type is correct."""
        self.assertEqual(DataStore.item_type(), ItemInfo.item_type())


class TestDataStoreWithToolbox(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Overridden method. Runs once before all tests in this class."""
        try:
            cls.app = QApplication().processEvents()
        except RuntimeError:
            pass
        logging.basicConfig(
            stream=sys.stderr,
            level=logging.DEBUG,
            format="%(asctime)s %(levelname)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )

    def setUp(self):
        """Set up."""
        self._temp_dir = TemporaryDirectory()
        self._toolbox = create_toolboxui_with_project(self._temp_dir.name)
        factory = DataStoreFactory()
        item_dict = {"type": "Data Store", "description": "", "x": 0, "y": 0, "url": None}
        self._project = self._toolbox.project()
        with mock.patch("spine_items.data_store.data_store.QMenu"):
            self.ds = factory.make_item("DS", item_dict, self._toolbox, self._project)
        self._project.add_item(self.ds)
        self._properties_widget = mock_finish_project_item_construction(factory, self.ds, self._toolbox)
        self.ds_properties_ui = self.ds._properties_ui

    def tearDown(self):
        """Overridden method. Runs after each test.
        Use this to free resources after a test if needed.
        """
        clean_up_toolbox(self._toolbox)
        ds_db_path = os.path.join(self.ds.data_dir, "DS.sqlite")
        if os.path.exists(ds_db_path):
            try:
                os.remove(ds_db_path)
            except OSError as os_e:
                logging.error("Failed to remove %s. Error: %s", ds_db_path, os_e)
        temp_db_path = os.path.join(self.ds.data_dir, "temp_db.sqlite")
        if os.path.exists(temp_db_path):
            try:
                os.remove(temp_db_path)
            except OSError as os_e:
                logging.error("Failed to remove %s. Error: %s", temp_db_path, os_e)
        gc.collect()
        self._temp_dir.cleanup()

    def create_temp_db(self):
        """Let's create a real db to more easily test complicated stuff (such as opening a tree view)."""
        temp_db_path = os.path.join(self.ds.data_dir, "temp_db.sqlite")
        sqlite_url = "sqlite:///" + temp_db_path
        create_new_spine_database(sqlite_url)
        gc.collect()
        return temp_db_path

    def test_rename(self):
        """Tests renaming a Data Store with an existing sqlite db in its data_dir."""
        temp_path = self.create_temp_db()
        url = {"dialect": "sqlite", "database": temp_path}
        self.ds._url = self.ds.parse_url(url)
        self.ds.activate()
        # Check that DS is connected to an existing DS.sqlite file that is in data_dir
        url = self.ds_properties_ui.url_selector_widget.url_dict()
        self.assertEqual(url["dialect"], "sqlite")
        self.assertTrue(
            os.path.normcase(url["database"]), os.path.normcase(os.path.join(self.ds.data_dir, "temp_db.sqlite"))
        )
        self.assertTrue(os.path.exists(url["database"]))
        expected_name = "ABC"
        expected_short_name = "abc"
        expected_data_dir = os.path.join(self._project.items_dir, expected_short_name)
        self.assertTrue(self.ds.rename(expected_name, ""))  # Do rename
        # Check name
        self.assertEqual(expected_name, self.ds.name)  # item name
        self.assertEqual(expected_name, self.ds.get_icon().name())  # name item on Design View
        # Check data_dir and logs_dir
        self.assertEqual(expected_data_dir, self.ds.data_dir)  # Check data dir
        # Check that the database path in properties has been updated
        expected_db_path = os.path.normcase(os.path.join(expected_data_dir, "temp_db.sqlite"))
        url = self.ds_properties_ui.url_selector_widget.url_dict()
        self.assertEqual(os.path.normcase(url["database"]), expected_db_path)
        # Check that the db file has actually been moved
        self.assertTrue(os.path.exists(url["database"]))

    def test_dirty_db_notification(self):
        temp_path = self.create_temp_db()
        url = {"dialect": "sqlite", "database": temp_path}
        self.ds._url = self.ds.parse_url(url)
        self.ds.activate()
        # Test that there are no notifications
        self.ds._check_notifications()
        self.assertEqual([], self.ds.get_icon().exclamation_icon._notifications)
        # Check that there is a warning about uncommitted changes
        db_map = self.ds.get_db_map()
        self._toolbox.db_mngr.add_items("entity_class", {db_map: [{"name": "my_object_class"}]})
        self.assertEqual(
            [f"{self.ds.name} has uncommitted changes"], self.ds.get_icon().exclamation_icon._notifications
        )
        # Check that the warning disappears after committing the changes
        self._toolbox.db_mngr.commit_session("Added entity classes", db_map)
        self.assertEqual([], self.ds.get_icon().exclamation_icon._notifications)


# noinspection PyUnusedLocal
class TestDataStoreWithMockToolbox(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Overridden method. Runs once before all tests in this class."""
        try:
            cls.app = QApplication().processEvents()
        except RuntimeError:
            pass
        logging.basicConfig(
            stream=sys.stderr,
            level=logging.DEBUG,
            format="%(asctime)s %(levelname)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )

    def setUp(self):
        """Set up."""
        self.toolbox = create_mock_toolbox()
        factory = DataStoreFactory()
        item_dict = {"type": "Data Store", "description": "", "x": 0, "y": 0, "url": None}
        self._temp_dir = TemporaryDirectory()
        self.project = create_mock_project(self._temp_dir.name)
        self.toolbox.project.return_value = self.project
        with mock.patch("spine_items.data_store.data_store.QMenu"):
            self.ds = factory.make_item("DS", item_dict, self.toolbox, self.project)
        self.project.get_item.return_value = self.ds
        self._properties_widget = mock_finish_project_item_construction(factory, self.ds, self.toolbox)
        self.ds_properties_ui = self.ds._properties_ui

    def tearDown(self):
        """Overridden method. Runs after each test.
        Use this to free resources after a test if needed.
        """
        ds_db_path = os.path.join(self.ds.data_dir, "DS.sqlite")
        temp_db_path = os.path.join(self.ds.data_dir, "temp_db.sqlite")
        self.ds.tear_down()
        if os.path.exists(ds_db_path):
            try:
                os.remove(ds_db_path)
            except OSError as os_e:
                logging.error("Failed to remove %s. Error: %s", ds_db_path, os_e)
        if os.path.exists(temp_db_path):
            try:
                os.remove(temp_db_path)
            except OSError as os_e:
                logging.error("Failed to remove %s. Error: %s", temp_db_path, os_e)
        gc.collect()
        self._temp_dir.cleanup()

    def create_temp_db(self):
        """Let's create a real db to more easily test complicated stuff (such as opening a tree view)."""
        temp_db_path = os.path.normcase(os.path.join(self.ds.data_dir, "temp_db.sqlite"))
        sqlite_url = "sqlite:///" + temp_db_path
        create_new_spine_database(sqlite_url)
        return temp_db_path

    def test_item_dict(self):
        """Tests Item dictionary creation."""
        d = self.ds.item_dict()
        a = ["type", "description", "x", "y", "url"]
        url_keys = ["dialect", "username", "password", "host", "port", "database"]
        for k in a:
            self.assertTrue(k in d, f"Key '{k}' not in dict {d}")
            if k == "url":
                for url_key in url_keys:
                    self.assertTrue(url_key in d[k], f"Key '{url_key}' not in dict {d[k]}")

    def test_create_new_empty_spine_database(self):
        """Test that an open file dialog is shown when clicking on 'New Spine db tool button'
        with an empty Data Store.
        """
        self.ds.activate()
        url = self.ds_properties_ui.url_selector_widget.url_dict()
        self.assertEqual(url["dialect"], "sqlite")
        self.assertEqual(url["database"], "")
        # Click New Spine db button
        self.toolbox.db_mngr = mock.MagicMock()
        with mock.patch("spine_items.data_store.data_store.QFileDialog") as file_dialog:
            file_dialog.getSaveFileName.return_value = ("", "")
            self.ds_properties_ui.pushButton_create_new_spine_db.click()
            file_dialog.getSaveFileName.assert_called_once()
        url = self.ds_properties_ui.url_selector_widget.url_dict()
        self.assertEqual(url["dialect"], "sqlite")
        self.assertEqual(url["database"], "")
        self.toolbox.db_mngr.create_new_spine_database.assert_not_called()

    def test_create_new_empty_spine_database2(self):
        """Test that a new Spine database is created when clicking on 'New Spine db' tool button
        with a Data Store that already has an .sqlite db.
        """
        temp_path = self.create_temp_db()
        # Connect to an existing .sqlite db
        url = {"dialect": "sqlite", "database": temp_path}
        self.ds._url = self.ds.parse_url(url)
        self.ds.activate()  # This loads the url into properties UI widgets
        # DS should now have "sqlite" selected in the combobox
        url = self.ds_properties_ui.url_selector_widget.url_dict()
        self.assertEqual(url["dialect"], "sqlite")
        self.assertEqual(url["database"], temp_path)
        self.assertTrue(os.path.exists(url["database"]))
        # Click New Spine db button. Creates the sqlite file.
        with mock.patch("spine_items.data_store.data_store.QFileDialog") as file_dialog:
            file_dialog.getSaveFileName.return_value = (temp_path, "*.sqlite")
            with signal_waiter(self.ds_properties_ui.pushButton_create_new_spine_db.clicked) as waiter:
                self.ds_properties_ui.pushButton_create_new_spine_db.click()
                waiter.wait()
            file_dialog.getSaveFileName.assert_called_once()
        url = self.ds_properties_ui.url_selector_widget.url_dict()
        self.assertEqual(url["database"], temp_path)
        self.assertEqual(url["dialect"], "sqlite")
        self.assertTrue(os.path.exists(url["database"]))
        self.toolbox.db_mngr.create_new_spine_database.assert_called_once()

    def test_new_database_is_created_before_advertising_resources(self):
        self.ds.activate()
        database_file_path = os.path.join(self._temp_dir.name, "test_db.sqlite")
        self.ds_properties_ui.url_selector_widget.set_url({"dialect": "sqlite"})
        self.ds._update_url_from_properties()
        url = self.ds_properties_ui.url_selector_widget.url_dict()
        self.assertEqual(url["dialect"], "sqlite")
        calls_with_non_empty_resources = []
        failure_messages = []

        def check_database_exists(data_store, for_successors, non_empty_calls):
            if for_successors:
                resources = data_store.resources_for_direct_successors()
            else:
                resources = data_store.resources_for_direct_predecessors()
            if len(resources) == 0:
                return
            non_empty_calls.append(1)
            if len(resources) > 1:
                failure_messages.append("resources length is not equal to 1")
                return
            call_args = self.toolbox.db_mngr.create_new_spine_database.call_args
            if len(call_args[0]) != 2:
                failure_messages.append("call_args *args length is not equal to 2")
                return
            if str(call_args[0][0]) != resources[0].url:
                failure_messages.append("create_new_spine_database() called with wrong URL")

        with mock.patch("spine_items.data_store.data_store.QFileDialog") as file_dialog:
            Path(database_file_path).write_bytes(b"fake sqlite file")
            file_dialog.getSaveFileName.return_value = (database_file_path,)
            self.project.notify_resource_changes_to_successors.side_effect = lambda item: check_database_exists(
                item, True, calls_with_non_empty_resources
            )
            self.project.notify_resource_changes_to_predecessors.side_effect = lambda item: check_database_exists(
                item, False, calls_with_non_empty_resources
            )
            self.assertTrue(self.ds_properties_ui.pushButton_create_new_spine_db.isEnabled())
            self.ds_properties_ui.pushButton_create_new_spine_db.click()
        while not self.ds.is_url_validated():
            QApplication.processEvents()
        self.assertGreater(len(calls_with_non_empty_resources), 0)
        if failure_messages:
            self.fail(failure_messages[0])

    def test_save_and_restore_mysql_selections(self):
        """Test that MySQL selections are saved and restored when
        deactivating a Data Store and activating it again."""
        url = {"dialect": "mysql", "database": "sqlite:///mock_db.sqlite"}
        self.ds._url = self.ds.parse_url(url)  # Set a URL for the Data Store
        self.ds.activate()
        self.ds_properties_ui.url_selector_widget.set_url(
            {
                "dialect": "mysql",
                "host": "localhost",
                "port": "8080",
                "database": "foo",
                "username": "bar",
                "password": "s3cr3t",
            }
        )
        self.ds._update_url_from_properties()
        self.ds.deactivate()
        self.ds.activate()
        url = self.ds_properties_ui.url_selector_widget.url_dict()
        self.assertEqual(url["dialect"], "mysql")
        self.assertEqual(url["host"], "localhost")
        self.assertEqual(url["port"], "8080")
        self.assertEqual(url["database"], "foo")
        self.assertEqual(url["username"], "bar")
        self.assertEqual(url["password"], "s3cr3t")

    def test_copy_db_url_to_clipboard(self):
        """Test that the database url from current selections is copied to clipboard."""
        QApplication.clipboard().clear()
        temp_path = self.create_temp_db()
        url = {"dialect": "sqlite", "database": temp_path}
        self.ds._url = self.ds.parse_url(url)
        self.ds.activate()  # This loads the url into properties UI widgets
        self.ds_properties_ui.toolButton_copy_url.click()
        # noinspection PyArgumentList
        clipboard_text = QApplication.clipboard().text()
        expected_url = "sqlite:///" + os.path.normcase(os.path.join(self.ds.data_dir, "temp_db.sqlite"))
        self.assertEqual(expected_url, clipboard_text.strip())

    def test_open_db_editor1(self):
        """Test that selecting the 'sqlite' dialect, browsing to an existing db file,
        and pressing open form works as expected.
        """
        temp_db_path = self.create_temp_db()
        self.ds.activate()
        self.ds_properties_ui.url_selector_widget.set_url({"dialect": "sqlite", "database": temp_db_path})
        self.ds._update_url_from_properties()
        self.toolbox.db_mngr = mock.MagicMock()
        while not self.ds.is_url_validated():
            QApplication.processEvents()
        self.assertTrue(self.ds_properties_ui.pushButton_ds_open_editor.isEnabled())
        with mock.patch("spine_items.data_store.data_store.open_db_editor") as open_db_editor:
            self.ds_properties_ui.pushButton_ds_open_editor.click()
            sa_url = convert_to_sqlalchemy_url(self.ds.url(), "DS", logger=None)
            self.assertIsNotNone(sa_url)
            open_db_editor.assert_called_with([sa_url], self.toolbox.db_mngr, True)

    def test_open_db_editor2(self):
        """Test that selecting the 'sqlite' dialect, typing the path to an existing db file,
        and pressing open form works as expected.
        """
        temp_db_path = self.create_temp_db()
        self.ds.activate()
        self.ds_properties_ui.url_selector_widget.set_url({"dialect": "sqlite", "database": temp_db_path})
        self.ds._update_url_from_properties()
        self.toolbox.db_mngr = mock.MagicMock()
        while not self.ds.is_url_validated():
            QApplication.processEvents()
        self.assertTrue(self.ds_properties_ui.pushButton_ds_open_editor.isEnabled())
        with mock.patch("spine_items.data_store.data_store.open_db_editor") as open_db_editor:
            self.ds_properties_ui.pushButton_ds_open_editor.click()
            sa_url = convert_to_sqlalchemy_url(self.ds.url(), "DS", logger=None)
            self.assertIsNotNone(sa_url)
            open_db_editor.assert_called_with([sa_url], self.toolbox.db_mngr, True)

    def test_notify_destination(self):
        self.ds.logger.msg = mock.MagicMock()
        self.ds.logger.msg_warning = mock.MagicMock()
        source_item = mock.MagicMock()
        source_item.name = "source name"
        source_item.item_type = mock.MagicMock(return_value="Data Connection")
        self.ds.notify_destination(source_item)
        self.ds.logger.msg_warning.emit.assert_called_with(
            "Link established. Interaction between a "
            "<b>Data Connection</b> and a <b>Data Store</b> has not been implemented yet."
        )
        source_item.item_type = mock.MagicMock(return_value="Importer")
        self.ds.notify_destination(source_item)
        self.ds.logger.msg.emit.assert_called_with(
            "Link established. "
            "Mapped data generated by <b>source name</b> will be imported in <b>DS</b> when executing."
        )
        source_item.item_type = mock.MagicMock(return_value="Tool")
        self.ds.notify_destination(source_item)
        self.ds.logger.msg.emit.assert_called_with("Link established")
        source_item.item_type = mock.MagicMock(return_value="View")
        self.ds.notify_destination(source_item)
        self.ds.logger.msg_warning.emit.assert_called_with(
            "Link established. Interaction between a "
            "<b>View</b> and a <b>Data Store</b> has not been implemented yet."
        )

    def test_do_update_url_uses_filterable_resources_when_replacing_them(self):
        database_1 = os.path.normcase(os.path.join(self._temp_dir.name, "db1.sqlite"))
        Path(database_1).write_bytes(b"empty fake db")
        self.ds.do_update_url(dialect="sqlite", database=database_1)
        self.project.notify_resource_changes_to_predecessors.assert_called_once_with(self.ds)
        self.project.notify_resource_changes_to_successors.assert_called_once_with(self.ds)
        while not self.ds.is_url_validated():
            QApplication.processEvents()
        database_2 = os.path.normcase(os.path.join(self._temp_dir.name, "db2.sqlite"))
        Path(database_2).write_bytes(b"empty fake db")
        self.ds.do_update_url(dialect="sqlite", database=database_2)
        while not self.ds.is_url_validated():
            QApplication.processEvents()
        expected_old_resources = [
            database_resource(
                self.ds.name, "sqlite:///" + database_1, label=database_label(self.ds.name), filterable=True
            )
        ]
        expected_new_resources = [
            database_resource(
                self.ds.name, "sqlite:///" + database_2, label=database_label(self.ds.name), filterable=True
            )
        ]
        self.project.notify_resource_replacement_to_predecessors.assert_called_once_with(
            self.ds, expected_old_resources, expected_new_resources
        )
        self.project.notify_resource_replacement_to_successors.assert_called_once_with(
            self.ds, expected_old_resources, expected_new_resources
        )


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains :class:`ClassTreeWidget`."""
import pickle
from PySide6.QtCore import QMimeData
from PySide6.QtWidgets import QMessageBox, QTreeWidget, QTreeWidgetItem
from spinedb_api import DatabaseMapping, SpineDBAPIError
from .drop_target_table import DROP_MIME_TYPE


class ClassTreeWidget(QTreeWidget):
    """A tree widget with drag capabilities to show entity classes."""

    def mimeData(self, items):
        mime_data = QMimeData()
        classes = [item.text(0) for item in items]
        mime_data.setData(DROP_MIME_TYPE, pickle.dumps(classes))
        return mime_data

    def load_data(self, url):
        """
        Loads parameter data from given URL.

        Args:
            url (str): database URL
        """
        classes = []
        try:
            with DatabaseMapping(url) as db_map:
                try:
                    for class_row in db_map.query(db_map.entity_class_sq):
                        classes.append(class_row.name)
                except SpineDBAPIError as error:
                    QMessageBox.information(
                        self, "Error while reading database", f"Could not read from database {url}:\n{error}"
                    )
        except SpineDBAPIError as error:
            QMessageBox.information(self, "Error while opening database", f"Could not open database {url}:\n{error}")
            return
        self.clear()
        for class_name in classes:
            self.addTopLevelItem(QTreeWidgetItem([class_name]))



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for the ``import_editor_window`` module."""
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from PySide6.QtWidgets import QApplication, QDialog
from spine_items.importer.widgets.import_editor_window import ImportEditorWindow
from spinedb_api.spine_io.importers.sqlalchemy_reader import SQLAlchemyReader
from tests.mock_helpers import clean_up_toolbox, create_toolboxui_with_project


class TestIsUrl(unittest.TestCase):
    def test_url_is_url(self):
        self.assertTrue(ImportEditorWindow._is_url("sqlite:///C:\\data\\db.sqlite"))
        self.assertTrue(ImportEditorWindow._is_url("postgresql+psycopg://spuser:s3cr3t@server.com/db"))

    def test_non_url_is_not_url(self):
        self.assertFalse(ImportEditorWindow._is_url("C:\\path\\to\\file"))
        self.assertFalse(ImportEditorWindow._is_url("/unix/style/path"))


class TestImportEditorWindow(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        self._temp_dir = TemporaryDirectory()
        self._toolbox = create_toolboxui_with_project(self._temp_dir.name)

    def tearDown(self):
        clean_up_toolbox(self._toolbox)
        self._temp_dir.cleanup()

    def test_get_reader_selects_sql_alchemy_reader_when_source_is_url(self):
        with mock.patch("spine_items.importer.widgets.import_editor_window.QDialog.exec") as exec_dialog:
            exec_dialog.return_value = QDialog.DialogCode.Accepted
            editor = ImportEditorWindow(self._toolbox, None)
            QApplication.processEvents()  # Let QTimer call ImportEditorWindow.start_ui()
            exec_dialog.assert_called_once()
            exec_dialog.reset_mock()
            reader = editor._get_reader("mysql://server.com/db")
            exec_dialog.assert_called_once()
            editor.close()
        self.assertIs(reader, SQLAlchemyReader)


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

""" Contains :class:`SpecificationEditorWindow`. """
from copy import deepcopy
import json
from PySide6.QtCore import QItemSelectionModel, QMimeData, QModelIndex, QPoint, Qt, Signal, Slot
from PySide6.QtGui import QAction, QKeySequence
from PySide6.QtWidgets import QApplication, QHeaderView, QMenu
from spinedb_api.export_mapping import (
    alternative_export,
    entity_dimension_parameter_default_value_export,
    entity_dimension_parameter_value_export,
    entity_export,
    entity_group_export,
    entity_parameter_default_value_export,
    entity_parameter_value_export,
    parameter_value_list_export,
    scenario_alternative_export,
    scenario_export,
)
from spinedb_api.export_mapping.export_mapping import FixedValueMapping, Position
from spinedb_api.export_mapping.group_functions import (
    GROUP_FUNCTION_DISPLAY_NAMES,
    NoGroup,
    group_function_display_from_name,
    group_function_name_from_display,
)
from spinedb_api.export_mapping.settings import (
    set_entity_dimensions,
    set_parameter_default_value_dimensions,
    set_parameter_dimensions,
)
from spinedb_api.mapping import unflatten
from spinetoolbox.helpers import SealCommand
from spinetoolbox.project_item.specification_editor_window import SpecificationEditorWindowBase
from ...widgets import combo_box_width
from ..commands import (
    ChangeWriteOrder,
    ClearFixedTableName,
    CompactMapping,
    DisableAllMappings,
    EnableAllMappings,
    NewMapping,
    RemoveMapping,
    RenameMapping,
    SetAlwaysExportHeader,
    SetExportFormat,
    SetFixedTableName,
    SetGroupFunction,
    SetHighlightDimension,
    SetMapping,
    SetMappingEnabled,
    SetMappingType,
    SetUseFixedTableNameFlag,
)
from ..mvcmodels.mapping_editor_table_model import POSITION_DISPLAY_TEXT, EditorColumn, MappingEditorTableModel
from ..mvcmodels.mappings_table_model import MappingsTableModel
from ..mvcmodels.mappings_table_proxy import MappingsTableProxy
from ..specification import MappingSpecification, MappingType, OutputFormat, Specification
from .filter_edit_delegate import FilterEditDelegate
from .position_edit_delegate import PositionEditDelegate
from .preview_updater import PreviewUpdater

mapping_type_to_combo_box_label = {
    MappingType.alternatives: "Alternative",
    MappingType.entities: "Entity class",
    MappingType.entity_groups: "Entity group",
    MappingType.entity_parameter_default_values: "Entity class",
    MappingType.entity_parameter_values: "Entity class",
    MappingType.parameter_value_lists: "Parameter value list",
    MappingType.entity_dimension_parameter_default_values: "Entity class with dimension parameter",
    MappingType.entity_dimension_parameter_values: "Entity class with dimension parameter",
    MappingType.scenario_alternatives: "Scenario alternative",
    MappingType.scenarios: "Scenario",
}

mapping_type_to_parameter_type_label = {
    MappingType.alternatives: "None",
    MappingType.entities: "None",
    MappingType.entity_groups: "None",
    MappingType.entity_parameter_default_values: "Default value",
    MappingType.entity_parameter_values: "Value",
    MappingType.entity_dimension_parameter_default_values: "Default value",
    MappingType.entity_dimension_parameter_values: "Value",
    MappingType.parameter_value_lists: "None",
    MappingType.scenario_alternatives: "None",
    MappingType.scenarios: "None",
}


_MAPPINGS_MIME_TYPE = "application/spine_items-exportmappings"


class SpecificationEditorWindow(SpecificationEditorWindowBase):
    """Interface to edit exporter specifications."""

    current_mapping_about_to_change = Signal()
    current_mapping_changed = Signal()

    def __init__(self, toolbox, specification=None, item=None, url_model=None):
        """
        Args:
            toolbox (ToolboxUI): Toolbox main window
            specification (Specification, optional): exporter specification
            item (ProjectItem, optional): invoking project item, if window was opened from its properties tab
            url_model (QAbstractListModel): model that provides URLs of connected databases
        """
        super().__init__(toolbox, specification, item)
        if specification is None:
            output_format = self._sniff_export_format() if item is not None else OutputFormat.default()
        else:
            output_format = specification.output_format
        self._new_spec = (
            deepcopy(specification) if specification is not None else Specification(output_format=output_format)
        )
        self._mappings_table_model = MappingsTableModel(self._new_spec.mapping_specifications(), self)
        self._mappings_table_model.dataChanged.connect(self._update_ui_after_mapping_change)
        self._mappings_table_model.rename_requested.connect(self._rename_mapping)
        self._mappings_table_model.mapping_enabled_state_change_requested.connect(self._set_mapping_enabled)
        self._mappings_table_model.set_all_mappings_enabled_requested.connect(self._enable_disable_all_mappings)
        self._mappings_table_model.write_order_about_to_change.connect(
            lambda tr=True: self._expect_current_mapping_change(tr)
        )
        self._mappings_table_model.write_order_changed.connect(lambda tr=False: self._expect_current_mapping_change(tr))
        self._mapping_editor_model = MappingEditorTableModel("", None, self._undo_stack, self, self)
        self._mapping_editor_model.dataChanged.connect(self._validate_fixed_table_name)
        self._sort_mappings_table_model = MappingsTableProxy(self)
        self._sort_mappings_table_model.setSortCaseSensitivity(Qt.CaseInsensitive)
        self._sort_mappings_table_model.setSourceModel(self._mappings_table_model)
        self._mappings_table_model.rowsInserted.connect(self._select_inserted_row)
        self._mappings_table_model.rowsRemoved.connect(self._check_for_empty_mappings_list)
        self._ui.export_format_combo_box.addItems([output_format.value for output_format in OutputFormat])
        self._ui.export_format_combo_box.setCurrentText(self._new_spec.output_format.value)
        self._ui.export_format_combo_box.currentTextChanged.connect(self._change_format)
        self._add_mapping_action = QAction("Add Mapping", self)
        self._add_mapping_action.triggered.connect(self._new_mapping)
        self._ui.add_mapping_button.clicked.connect(self._add_mapping_action.trigger)
        self._remove_mapping_action = QAction("Remove mappings", self)
        self._remove_mapping_action.setShortcut(QKeySequence(QKeySequence.Delete))
        self._remove_mapping_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self._remove_mapping_action.triggered.connect(self._delete_mapping)
        self._copy_mappings_action = QAction("Copy", self)
        self._copy_mappings_action.setShortcut(QKeySequence(QKeySequence.Copy))
        self._copy_mappings_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self._copy_mappings_action.triggered.connect(self._copy_mappings_to_clipboard)
        self._paste_mappings_action = QAction("Paste", self)
        self._paste_mappings_action.setShortcut(QKeySequence(QKeySequence.Paste))
        self._paste_mappings_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self._paste_mappings_action.triggered.connect(self._paste_mappings_from_clipboard)
        self._duplicate_mappings_action = QAction("Duplicate", self)
        self._duplicate_mappings_action.triggered.connect(self._duplicate_mappings)
        self._ui.remove_mapping_button.clicked.connect(self._remove_mapping_action.trigger)
        self._toggle_all_enabled_action = QAction("Toggle all mappings enabled", self)
        self._toggle_all_enabled_action.triggered.connect(self._toggle_all_enabled)
        self._ui.toggle_enabled_button.clicked.connect(self._toggle_all_enabled_action.trigger)
        self._write_earlier_action = QAction("Write earlier", self)
        self._write_earlier_action.triggered.connect(self._write_earlier)
        self._ui.write_earlier_button.clicked.connect(self._write_earlier_action.trigger)
        self._write_later_action = QAction("Write later", self)
        self._write_later_action.triggered.connect(self._write_later)
        self._ui.write_later_button.clicked.connect(self._write_later_action.trigger)
        self._mappings_table_menu = self._make_mappings_table_menu()
        self._ui.mappings_table.addAction(self._add_mapping_action)
        self._ui.mappings_table.addAction(self._remove_mapping_action)
        self._ui.mappings_table.addAction(self._copy_mappings_action)
        self._ui.mappings_table.addAction(self._paste_mappings_action)
        self._ui.mappings_table.setModel(self._sort_mappings_table_model)
        self._ui.mappings_table.setSortingEnabled(True)
        self._ui.mappings_table.sortByColumn(0, Qt.AscendingOrder)
        self._ui.mappings_table.customContextMenuRequested.connect(self._show_mappings_table_context_menu)
        self._expect_current_mapping_change(False)
        self._ui.mappings_table.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        self._ui.group_fn_combo_box.addItems(GROUP_FUNCTION_DISPLAY_NAMES)
        self._ui.item_type_combo_box.currentTextChanged.connect(self._change_mapping_type)
        self._ui.parameter_type_combo_box.currentTextChanged.connect(self._change_mapping_type)
        self._ui.parameter_dimensions_spin_box.valueChanged.connect(self._change_parameter_dimensions)
        self._ui.always_export_header_check_box.stateChanged.connect(self._change_always_export_header)
        self._ui.entity_dimensions_spin_box.valueChanged.connect(self._change_entity_dimensions)
        self._ui.highlight_dimension_spin_box.valueChanged.connect(self._change_highlight_dimension)
        self._ui.fix_table_name_check_box.stateChanged.connect(self._change_fix_table_name_flag)
        self._ui.fix_table_name_line_edit.textEdited.connect(self._change_fix_table_name)
        self._ui.fix_table_name_line_edit.editingFinished.connect(self._finish_editing_fix_table_name)
        self._ui.group_fn_combo_box.currentTextChanged.connect(self._change_root_mapping_group_fn)
        self._compact_mapping_action = QAction("Compact mapping", self)
        self._compact_mapping_action.triggered.connect(self._compact_mapping)
        self._ui.compact_button.clicked.connect(self._compact_mapping_action.trigger)
        self._ui.mapping_table_view.setModel(self._mapping_editor_model)
        self._position_edit_delegate = PositionEditDelegate(self)
        self._ui.mapping_table_view.setItemDelegateForColumn(EditorColumn.POSITION, self._position_edit_delegate)
        self._filter_edit_delegate = FilterEditDelegate(self)
        self._ui.mapping_table_view.setItemDelegateForColumn(EditorColumn.FILTER, self._filter_edit_delegate)
        table_header = self._ui.mapping_table_view.horizontalHeader()
        table_header.setSectionResizeMode(QHeaderView.ResizeMode.ResizeToContents)
        table_header.setSectionResizeMode(EditorColumn.POSITION, QHeaderView.ResizeMode.Fixed)
        table_header.resizeSection(EditorColumn.POSITION, combo_box_width(self, POSITION_DISPLAY_TEXT.values()))
        self._enable_mapping_specification_editing()
        if specification is None:
            self._mappings_table_model.extend(_new_mapping_specification(MappingType.entities))
        self._preview_updater = PreviewUpdater(
            self,
            self._ui,
            url_model,
            self._mappings_table_model,
            self._sort_mappings_table_model,
            self._mapping_editor_model,
            self._toolbox.project().project_dir,
        )
        self._ui.mappings_table.setCurrentIndex(self._sort_mappings_table_model.index(0, 0))

    @property
    def settings_group(self):
        return "exportSpecificationEditorWindow"

    def _make_ui(self):
        from ..ui.specification_editor import Ui_MainWindow  # pylint: disable=import-outside-toplevel

        return Ui_MainWindow()

    def _make_new_specification(self, spec_name):
        """See base class."""
        description = self._spec_toolbar.description()
        mapping_specification = deepcopy(self._new_spec.mapping_specifications())
        output_format = self._new_spec.output_format
        return Specification(spec_name, description, mapping_specification, output_format)

    @Slot(str)
    def _change_format(self, current):
        """
        Pusher ``SetExportFormat`` command to undo stack.

        Args:
            current (str): new export format
        """
        output_format = OutputFormat(current)
        self._undo_stack.push(SetExportFormat(self, output_format, self._new_spec.output_format))

    def set_export_format_silently(self, export_format):
        """
        Sets export format.

        Args:
            export_format (ExportFormat): new format
        """
        self._new_spec.output_format = export_format
        if export_format.value != self._ui.export_format_combo_box.currentText():
            self._ui.export_format_combo_box.currentTextChanged.disconnect(self._change_format)
            self._ui.export_format_combo_box.setCurrentText(export_format.value)
            self._ui.export_format_combo_box.currentTextChanged.connect(self._change_format)

    def _sniff_export_format(self):
        """Tries to guess the export file format from user given export label.

        Returns:
            OutputFormat: export file format
        """
        if self.item.has_out_url():
            return OutputFormat.SQL
        out_labels = self.item.get_out_labels()
        for label in out_labels:
            _, separator, extension = label.rpartition(".")
            if not separator:
                continue
            output_format = OutputFormat.output_format_from_extension(extension)
            if output_format is None:
                return OutputFormat.default()
            return output_format if output_format is not None else OutputFormat.default()
        return OutputFormat.default()

    def show_on_table(self, mapping_name):
        """
        Changes the current mapping.

        Args:
            mapping_name (str): mapping's name
        """
        index = self._sort_mappings_table_model.mapFromSource(self._mappings_table_model.index_of(mapping_name))
        self._ui.mappings_table.selectionModel().setCurrentIndex(index, QItemSelectionModel.ClearAndSelect)

    @Slot(QModelIndex, QModelIndex)
    def change_current_mapping(self, current, previous):
        """
        Changes the current mapping.

        Args:
            current (QModelIndex): current mapping on mapping name list
            previous (QModelIndex): previous mapping name
        """
        if current.row() == previous.row():
            return
        if not current.isValid():
            self._mapping_editor_model.set_mapping("", None)
            self._ui.mapping_table_view.setEnabled(False)
            return
        self.current_mapping_about_to_change.emit()
        current = self._sort_mappings_table_model.mapToSource(current)
        mapping_type = MappingType(current.data(MappingsTableModel.MAPPING_TYPE_ROLE))
        self._set_mapping_type_silently(mapping_type_to_combo_box_label[mapping_type])
        self._set_parameter_type_silently(mapping_type_to_parameter_type_label[mapping_type])
        self._set_always_export_header_silently(current.data(MappingsTableModel.ALWAYS_EXPORT_HEADER_ROLE))
        if mapping_type in {
            MappingType.entities,
            MappingType.entity_parameter_values,
            MappingType.entity_parameter_default_values,
            MappingType.entity_dimension_parameter_values,
            MappingType.entity_dimension_parameter_default_values,
        }:
            self._ui.entity_dimensions_spin_box.setMinimum(0)
            self._set_entity_dimensions_silently(current.data(MappingsTableModel.ENTITY_DIMENSIONS_ROLE))
            if mapping_type in {
                MappingType.entity_dimension_parameter_values,
                MappingType.entity_dimension_parameter_default_values,
            }:
                self._ui.entity_dimensions_spin_box.setMinimum(1)
        else:
            self._set_entity_dimensions_silently(0)
            self._ui.entity_dimensions_spin_box.setMinimum(0)
        if mapping_type in {
            MappingType.entity_dimension_parameter_values,
            MappingType.entity_dimension_parameter_default_values,
        }:
            self._set_highlight_dimension_silently(current.data(MappingsTableModel.HIGHLIGHT_POSITION_ROLE))
            self._ui.highlight_dimension_spin_box.setMaximum(current.data(MappingsTableModel.ENTITY_DIMENSIONS_ROLE))
        else:
            self._set_highlight_dimension_silently(0)
        if mapping_type in {
            MappingType.entity_parameter_values,
            MappingType.entity_parameter_default_values,
            MappingType.entity_dimension_parameter_values,
            MappingType.entity_dimension_parameter_default_values,
        }:
            self._set_parameter_dimensions_silently(current.data(MappingsTableModel.PARAMETER_DIMENSIONS_ROLE))
        else:
            self._set_parameter_dimensions_silently(0)
        root_mapping = current.data(MappingsTableModel.MAPPING_ROOT_ROLE)
        self._set_use_fixed_table_name_flag_silently(current.data(MappingsTableModel.USE_FIXED_TABLE_NAME_FLAG_ROLE))
        self._set_fixed_table_name_silently(current.data(MappingsTableModel.FIXED_TABLE_NAME_ROLE))
        if any(
            m.position == Position.table_name for m in root_mapping.flatten() if not isinstance(m, FixedValueMapping)
        ):
            self._ui.fix_table_name_check_box.setEnabled(False)
            self._ui.fix_table_name_line_edit.setEnabled(False)
        self._set_group_fn_silently(group_function_display_from_name(current.data(MappingsTableModel.GROUP_FN_ROLE)))
        mapping_name = self._mappings_table_model.index(current.row(), 0).data()
        self._mapping_editor_model.set_mapping(mapping_name, root_mapping)
        self._enable_entity_controls()
        self._enable_parameter_controls()
        self._ui.mapping_table_view.setEnabled(True)
        self.current_mapping_changed.emit()

    @Slot(bool)
    def _expect_current_mapping_change(self, expect_change):
        """Connects and disconnects signals depending on if current mapping is expected to change.

        Args:
            expect_change (bool): True to expect changes
        """
        if expect_change:
            self._ui.mappings_table.selectionModel().currentChanged.disconnect(self.change_current_mapping)
        else:
            self._ui.mappings_table.selectionModel().currentChanged.connect(self.change_current_mapping)

    @Slot(QModelIndex, QModelIndex, list)
    def _update_ui_after_mapping_change(self, top_left, bottom_right, roles):
        """
        Makes sure we show the correct mapping data on the window.
        Receiver for ``self._mappings_table_model``'s dataChanged signal

        Args:
            top_left (QModelIndex): top index of modified mappings
            bottom_right (QModelIndex): bottom index of modified mappings
            roles (list of int):
        """
        if Qt.ItemDataRole.DisplayRole in roles:
            self._sort_mappings_table_model.invalidate()
        if max(roles) < Qt.ItemDataRole.UserRole:
            return
        sorted_index = self._sort_mappings_table_model.mapFromSource(top_left)
        if sorted_index != self._ui.mappings_table.currentIndex():
            self._ui.mappings_table.selectionModel().setCurrentIndex(sorted_index, QItemSelectionModel.ClearAndSelect)
            return
        enable_controls = False
        if MappingsTableModel.MAPPING_ROOT_ROLE in roles:
            root_mapping = top_left.data(MappingsTableModel.MAPPING_ROOT_ROLE)
            self._mapping_editor_model.set_mapping(top_left.data(Qt.ItemDataRole.DisplayRole), root_mapping)
            self._set_entity_dimensions_silently(top_left.data(MappingsTableModel.ENTITY_DIMENSIONS_ROLE))
            self._set_parameter_dimensions_silently(top_left.data(MappingsTableModel.PARAMETER_DIMENSIONS_ROLE))
            enable_controls = True
        if MappingsTableModel.MAPPING_TYPE_ROLE in roles:
            mapping_type = MappingType(top_left.data(MappingsTableModel.MAPPING_TYPE_ROLE))
            self._set_mapping_type_silently(mapping_type_to_combo_box_label[mapping_type])
            self._set_parameter_type_silently(mapping_type_to_parameter_type_label[mapping_type])
            if mapping_type in (
                MappingType.entity_dimension_parameter_default_values,
                MappingType.entity_dimension_parameter_values,
            ):
                self._ui.entity_dimensions_spin_box.setMinimum(1)
            else:
                self._ui.entity_dimensions_spin_box.setMinimum(0)
            enable_controls = True
        if MappingsTableModel.ALWAYS_EXPORT_HEADER_ROLE in roles:
            self._set_always_export_header_silently(top_left.data(MappingsTableModel.ALWAYS_EXPORT_HEADER_ROLE))
        if MappingsTableModel.USE_FIXED_TABLE_NAME_FLAG_ROLE in roles:
            self._set_use_fixed_table_name_flag_silently(
                top_left.data(MappingsTableModel.USE_FIXED_TABLE_NAME_FLAG_ROLE)
            )
        if {MappingsTableModel.MAPPING_ROOT_ROLE, MappingsTableModel.FIXED_TABLE_NAME_ROLE} & set(roles):
            self._set_fixed_table_name_silently(top_left.data(MappingsTableModel.FIXED_TABLE_NAME_ROLE))
        if MappingsTableModel.GROUP_FN_ROLE in roles:
            self._set_group_fn_silently(
                group_function_display_from_name(top_left.data(MappingsTableModel.GROUP_FN_ROLE))
            )
        if MappingsTableModel.HIGHLIGHT_POSITION_ROLE in roles:
            self._set_highlight_dimension_silently(top_left.data(MappingsTableModel.HIGHLIGHT_POSITION_ROLE))
        if enable_controls:
            self._enable_entity_controls()
            self._enable_parameter_controls()

    @Slot(QPoint)
    def _show_mappings_table_context_menu(self, position):
        """Shows Mappings table context menu.

        Args:
            position (QPoint): requested menu position
        """
        self._mappings_table_menu.exec(self._ui.mappings_table.mapToGlobal(position))

    def _make_mappings_table_menu(self):
        """Creates context menu for Mappings table.

        Returns:
            QMenu: context menu
        """
        menu = QMenu(self)
        menu.addAction(self._add_mapping_action)
        menu.addAction(self._copy_mappings_action)
        menu.addAction(self._paste_mappings_action)
        menu.addAction(self._duplicate_mappings_action)
        menu.addAction(self._remove_mapping_action)
        return menu

    @Slot(bool)
    def _new_mapping(self, _=True):
        """Pushes an add mapping command to the undo stack."""
        type_ = MappingType.entities
        mapping_specification = _new_mapping_specification(type_)
        self._undo_stack.push(NewMapping(self._mappings_table_model, mapping_specification))
        self._sort_mappings_table_model.invalidate()

    @Slot(QModelIndex, int, int)
    def _select_inserted_row(self, parent_index, first_row, last_row):
        """
        Selects newly inserted mapping.

        Args:
            parent_index (QModelIndex): ignored
            first_row (int): index of first inserted row
            last_row (int): index of last inserted row
        """
        index = self._sort_mappings_table_model.index(first_row, 0)
        self._ui.mappings_table.selectionModel().setCurrentIndex(index, QItemSelectionModel.ClearAndSelect)
        self._enable_mapping_specification_editing()

    @Slot()
    def _delete_mapping(self):
        """Pushes remove mapping commands for selected mappings to undo stack."""
        selection_model = self._ui.mappings_table.selectionModel()
        if not selection_model.hasSelection():
            return
        indexes = [
            self._sort_mappings_table_model.mapToSource(index)
            for index in selection_model.selectedIndexes()
            if index.column() == 0
        ]
        if len(indexes) == 1:
            row = indexes[0].row()
            self._undo_stack.push(RemoveMapping(row, self._mappings_table_model))
        else:
            self._undo_stack.beginMacro("remove mappings")
            names = [index.data() for index in indexes]
            for name in names:
                row = self._mappings_table_model.index_of(name).row()
                self._undo_stack.push(RemoveMapping(row, self._mappings_table_model))
            self._undo_stack.endMacro()

    @Slot(bool)
    def _copy_mappings_to_clipboard(self, checked):
        """Copies selected mappings to system clipboard.

        Args:
            checked (bool): unused
        """
        selection_model = self._ui.mappings_table.selectionModel()
        indices = [self._sort_mappings_table_model.mapToSource(i) for i in selection_model.selectedIndexes()]
        if not indices:
            return
        specifications = []
        for index in indices:
            if index.column() != 0:
                continue
            specification = self._mappings_table_model.data(index, MappingsTableModel.MAPPING_SPECIFICATION_ROLE)
            serialized = specification.to_dict()
            serialized["name"] = self._mappings_table_model.data(index)
            specifications.append(serialized)
        clipboard = QApplication.clipboard()
        mime_data = QMimeData()
        mime_data.setData(_MAPPINGS_MIME_TYPE, bytes(json.dumps(specifications), "utf-8"))
        clipboard.setMimeData(mime_data)

    @Slot(bool)
    def _paste_mappings_from_clipboard(self, checked):
        """Pastes mappings from system clipboard.

        Args:
            checked (bool): unused
        """
        clipboard = QApplication.clipboard()
        mime_data = clipboard.mimeData()
        if not mime_data.hasFormat(_MAPPINGS_MIME_TYPE):
            return
        serialized = json.loads(str(mime_data.data(_MAPPINGS_MIME_TYPE), "utf-8"))
        names = [d.pop("name") for d in serialized]
        mappings = [MappingSpecification.from_dict(d) for d in serialized]
        command_text = "paste mapping" + ("s" if len(mappings) > 1 else "")
        self._undo_stack.beginMacro(command_text)
        for mapping, name in zip(mappings, names):
            self._undo_stack.push(NewMapping(self._mappings_table_model, mapping, name))
        self._undo_stack.endMacro()

    @Slot(bool)
    def _duplicate_mappings(self, checked):
        """Duplicates selected mappings.

        Args:
            checked (bool): unused
        """
        selection_model = self._ui.mappings_table.selectionModel()
        indices = [self._sort_mappings_table_model.mapToSource(i) for i in selection_model.selectedIndexes()]
        if not indices:
            return
        specifications = []
        names = []
        for index in indices:
            if index.column() != 0:
                continue
            specification = self._mappings_table_model.data(index, MappingsTableModel.MAPPING_SPECIFICATION_ROLE)
            specifications.append(deepcopy(specification))
            names.append(self._mappings_table_model.data(index))
        text = "duplicate mapping" + ("s" if len(specifications) > 1 else "")
        self._undo_stack.beginMacro(text)
        for name, specification in zip(names, specifications):
            self._undo_stack.push(NewMapping(self._mappings_table_model, specification, name))
        self._undo_stack.endMacro()

    @Slot()
    def _toggle_all_enabled(self):
        """Pushes a command that enables or disables all mappings to undo stack."""
        for row in range(self._mappings_table_model.rowCount()):
            if self._mappings_table_model.index(row, 0).data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Unchecked:
                self._undo_stack.push(EnableAllMappings(self._mappings_table_model))
                return
        self._undo_stack.push(DisableAllMappings(self._mappings_table_model))

    @Slot()
    def _write_earlier(self):
        """Pushes a command that modifies mapping's write order."""
        index = self._ui.mappings_table.selectionModel().currentIndex()
        row = self._sort_mappings_table_model.mapToSource(index).row()
        if row == 0:
            return
        self._undo_stack.push(ChangeWriteOrder(row, True, self._mappings_table_model))

    @Slot()
    def _write_later(self):
        """Pushes a command that modifies mapping's write order."""
        index = self._ui.mappings_table.selectionModel().currentIndex()
        row = self._sort_mappings_table_model.mapToSource(index).row()
        if row == self._mappings_table_model.rowCount() - 1:
            return
        self._undo_stack.push(ChangeWriteOrder(row, False, self._mappings_table_model))

    @Slot(QModelIndex, int, int)
    def _check_for_empty_mappings_list(self, parent_index, first_row, last_row):
        """
        Checks if mappings list has become empty and disables controls accordingly.

        Args:
            parent_index (QModelIndex): ignored
            first_row (int): first removed row
            last_row (int): last removed row
        """
        self._enable_mapping_specification_editing()

    @Slot(int, str)
    def _rename_mapping(self, row, new_name):
        """
        Pushes a ``RenameMapping`` command to undo stack.

        Args:
            row (int): row index in mapping list model
            new_name (str): mapping's new name
        """
        self._undo_stack.push(RenameMapping(row, self._mappings_table_model, new_name))
        self._sort_mappings_table_model.invalidate()

    @Slot(int)
    def _set_mapping_enabled(self, row):
        """
        Pushes a ``SetMappingEnabled`` command to undo stack.

        Args:
            row (int): row index in mapping list model
        """
        self._undo_stack.push(SetMappingEnabled(row, self._mappings_table_model))

    @Slot(bool)
    def _enable_disable_all_mappings(self, enabled):
        """
        Pushes a ``EnableAllMappings`` or ``DisableAllMappings`` command to undo stack.

        Args:
            enabled (bool): True to enable all mapping, False to disable
        """
        make_command = EnableAllMappings if enabled else DisableAllMappings
        self._undo_stack.push(make_command(self._mappings_table_model))

    @Slot(str)
    def _change_mapping_type(self, _):
        """
        Pushes ``SetMappingType`` command to undo stack.

        Args:
            _ (str): ignored
        """
        index = self._sort_mappings_table_model.mapToSource(self._ui.mappings_table.currentIndex())
        if not index.isValid():
            return
        type_label = self._ui.item_type_combo_box.currentText()
        parameter_type_label = self._ui.parameter_type_combo_box.currentText()
        if type_label == "Entity class":
            if parameter_type_label == "None":
                mapping_type = MappingType.entities
            elif parameter_type_label == "Value":
                mapping_type = MappingType.entity_parameter_values
            else:
                mapping_type = MappingType.entity_parameter_default_values
        elif type_label == "Entity class with dimension parameter":
            if parameter_type_label == "Default value":
                mapping_type = MappingType.entity_dimension_parameter_default_values
            else:
                mapping_type = MappingType.entity_dimension_parameter_values
        elif type_label == "Entity group":
            mapping_type = MappingType.entity_groups
        else:
            mapping_type = {
                "Alternative": MappingType.alternatives,
                "Scenario": MappingType.scenarios,
                "Scenario alternative": MappingType.scenario_alternatives,
                "Parameter value list": MappingType.parameter_value_lists,
            }[type_label]
        mapping_specification = _new_mapping_specification(mapping_type)
        if self._ui.fix_table_name_check_box.isChecked():
            mapping_specification.root = _add_fixed_table_name(mapping_specification.root)
            mapping_specification.root.value = self._ui.fix_table_name_line_edit.text()
        self._undo_stack.beginMacro("change mapping type")
        self._undo_stack.push(SetMappingType(index, mapping_type))
        self._undo_stack.push(SetMapping(index, mapping_specification.root))
        self._undo_stack.endMacro()

    def _set_mapping_type_silently(self, type_name):
        """
        Changes mapping type in the combo box without any other side effects.

        Args:
            type_name (str): mapping's type
        """
        if type_name == self._ui.item_type_combo_box.currentText():
            return
        self._ui.item_type_combo_box.currentTextChanged.disconnect(self._change_mapping_type)
        self._ui.item_type_combo_box.setCurrentText(type_name)
        self._ui.item_type_combo_box.currentTextChanged.connect(self._change_mapping_type)
        self._enable_entity_controls()
        self._enable_parameter_controls()

    def _set_parameter_type_silently(self, type_name):
        """
        Changes parameter type in the combo box without any side effects.

        Args:
            type_name (str): parameter type
        """
        if type_name == self._ui.parameter_type_combo_box.currentText():
            return
        self._ui.parameter_type_combo_box.currentTextChanged.disconnect(self._change_mapping_type)
        self._ui.parameter_type_combo_box.setCurrentText(type_name)
        self._ui.parameter_type_combo_box.currentTextChanged.connect(self._change_mapping_type)
        self._enable_parameter_controls()

    @Slot(int)
    def _change_always_export_header(self, checked):
        """Pushes corresponding command to undo stack.

        Args:
            checked (int): checked state
        """
        index = self._ui.mappings_table.currentIndex()
        self._undo_stack.push(SetAlwaysExportHeader(index, checked == Qt.CheckState.Checked.value))

    def _set_always_export_header_silently(self, always_export_header):
        """
        Changes the always export header check box  without any side effects.

        Args:
            always_export_header (bool): True to check the box, false otherwise
        """
        if always_export_header == self._ui.always_export_header_check_box.isChecked():
            return
        self._ui.always_export_header_check_box.stateChanged.disconnect(self._change_always_export_header)
        self._ui.always_export_header_check_box.setCheckState(
            Qt.CheckState.Checked if always_export_header else Qt.CheckState.Unchecked
        )
        self._ui.always_export_header_check_box.stateChanged.connect(self._change_always_export_header)

    def _change_parameter_dimensions(self, dimensions):
        """
        Pushes a command to undo stack.

        Args:
            dimensions (int): parameter dimensions
        """
        index = self._sort_mappings_table_model.mapToSource(self._ui.mappings_table.currentIndex())
        mapping = self._mappings_table_model.data(index, MappingsTableModel.MAPPING_ROOT_ROLE)
        modified = deepcopy(mapping)
        if self._mappings_table_model.data(index, MappingsTableModel.MAPPING_TYPE_ROLE) in (
            MappingType.entity_parameter_values.value,
            MappingType.entity_dimension_parameter_values.value,
        ):
            set_parameter_dimensions(modified, dimensions)
        else:
            set_parameter_default_value_dimensions(modified, dimensions)
        self._undo_stack.beginMacro("change parameter dimensions")
        self._undo_stack.push(SetMapping(index, modified))
        self._undo_stack.endMacro()

    def _set_parameter_dimensions_silently(self, dimensions):
        """
        Sets parameter dimensions spin box without emitting signals.

        Args:
            dimensions (int): parameter dimensions
        """
        self._ui.parameter_dimensions_spin_box.valueChanged.disconnect(self._change_parameter_dimensions)
        self._ui.parameter_dimensions_spin_box.setValue(dimensions)
        self._ui.parameter_dimensions_spin_box.valueChanged.connect(self._change_parameter_dimensions)

    @Slot(int)
    def _change_fix_table_name_flag(self, checked):
        """
        Pushes commands to undo stack to change enable/disable fixed table name.

        Args:
            checked (int): check box state
        """
        index = self._ui.mappings_table.currentIndex()
        if not index.isValid():
            return
        flag = checked == Qt.CheckState.Checked.value
        self._undo_stack.beginMacro(("check" if flag else "uncheck") + " fix table name checkbox")
        self._undo_stack.push(SetUseFixedTableNameFlag(index, flag))
        mapping = index.data(MappingsTableModel.MAPPING_ROOT_ROLE)
        mapping = deepcopy(mapping)
        mapping = _add_fixed_table_name(mapping) if flag else _remove_fixed_table_name(mapping)
        self._undo_stack.push(SetMapping(index, mapping))
        self._undo_stack.endMacro()

    @Slot(str)
    def _change_fix_table_name(self, name):
        """
        Pushes commands to undo stack to change fixed table name to text in corresponding line edit.

        Args:
            name (str): fixed table name
        """
        index = self._ui.mappings_table.currentIndex()
        if not index.isValid():
            return
        old_name = index.data(MappingsTableModel.FIXED_TABLE_NAME_ROLE)
        self._undo_stack.push(SetFixedTableName(index, old_name, name))

    @Slot()
    def _finish_editing_fix_table_name(self):
        """Seals the latest undo command if it changed the fixed table name."""
        self._undo_stack.push(SealCommand(SetFixedTableName.ID))

    def _set_use_fixed_table_name_flag_silently(self, flag):
        """
        Changes the fixed table name check box without touching the undo stack.

        Args:
            flag (bool): True if the box should be checked, False otherwise
        """
        self._ui.fix_table_name_line_edit.setEnabled(flag)
        if flag == self._ui.fix_table_name_check_box.isChecked():
            return
        self._ui.fix_table_name_check_box.stateChanged.disconnect(self._change_fix_table_name_flag)
        self._ui.fix_table_name_check_box.setCheckState(Qt.CheckState.Checked if flag else Qt.CheckState.Unchecked)
        self._ui.fix_table_name_check_box.stateChanged.connect(self._change_fix_table_name_flag)

    def _set_fixed_table_name_silently(self, name):
        """
        Changes the fixed table name line edit without touching the undo stack.

        Args:
            name (str): text to go in the line edit
        """
        if name == self._ui.fix_table_name_line_edit.text():
            return
        self._ui.fix_table_name_line_edit.editingFinished.disconnect(self._finish_editing_fix_table_name)
        self._ui.fix_table_name_line_edit.setText(name)
        self._ui.fix_table_name_line_edit.editingFinished.connect(self._finish_editing_fix_table_name)

    @Slot(QModelIndex, QModelIndex, list)
    def _validate_fixed_table_name(self, top_left, bottom_right, roles):
        """Checks if any mapping position is set to Table name and disables fixed table name accordingly.

        Args:
            top_left (QModelIndex): top left of changed mapping table model index
            bottom_right (QModelIndex): bottom right of changed mapping table model index
            roles (list of int): model roles
        """
        if (
            Qt.ItemDataRole.DisplayRole not in roles
            or top_left.column() > EditorColumn.POSITION
            or bottom_right.column() < EditorColumn.POSITION
        ):
            return
        for row in range(top_left.row(), bottom_right.row() + 1):
            mapping_item = self._mapping_editor_model.index(row, EditorColumn.POSITION).data(
                MappingEditorTableModel.MAPPING_ITEM_ROLE
            )
            if mapping_item.position == Position.table_name:
                if self._ui.fix_table_name_check_box.isChecked():
                    index = self._ui.mappings_table.currentIndex()
                    mapping_root = index.data(MappingsTableModel.MAPPING_ROOT_ROLE)
                    mapping_root = deepcopy(mapping_root)
                    mapping_root = _remove_fixed_table_name(mapping_root)
                    command = ClearFixedTableName(
                        self,
                        self._ui.fix_table_name_check_box.isChecked(),
                        self._ui.fix_table_name_line_edit.text(),
                        index,
                        mapping_root,
                    )
                    self._undo_stack.push(command)
                else:
                    self._ui.fix_table_name_check_box.setEnabled(False)
                    self._ui.fix_table_name_line_edit.setEnabled(False)
                return
        for row in range(top_left.model().rowCount()):
            mapping_item = self._mapping_editor_model.index(row, EditorColumn.POSITION).data(
                MappingEditorTableModel.MAPPING_ITEM_ROLE
            )
            if mapping_item.position == Position.table_name:
                self._ui.fix_table_name_check_box.setEnabled(False)
                self._ui.fix_table_name_line_edit.setEnabled(False)
                return
        self._ui.fix_table_name_check_box.setEnabled(True)

    def clear_fixed_table_name(self):
        """Clears and disable fixed table name widgets."""
        self._set_fixed_table_name_silently("")
        self._set_use_fixed_table_name_flag_silently(False)
        self._ui.fix_table_name_check_box.setEnabled(False)

    def enable_fixed_table_name(self, checked, name):
        """Enables and sets fixed table name widgets.

        Args:
            checked (bool): fixed table name checkbox state
            name (str): fixed table name
        """
        self._set_use_fixed_table_name_flag_silently(checked)
        self._ui.fix_table_name_check_box.setEnabled(True)
        self._set_fixed_table_name_silently(name)
        self._ui.fix_table_name_line_edit.setEnabled(True)

    @Slot(str)
    def _change_root_mapping_group_fn(self, text):
        """
        Pushes commands to undo stack to change group function of root mapping.

        Args:
            text (str): combo box text
        """
        index = self._sort_mappings_table_model.mapToSource(self._ui.mappings_table.currentIndex())
        if not index.isValid():
            return
        old = index.data(MappingsTableModel.GROUP_FN_ROLE)
        self._undo_stack.push(SetGroupFunction(index, old, group_function_name_from_display(text)))

    def _set_group_fn_silently(self, group_fn_display_name):
        """
        Sets group function in combo box without emitting signals.

        Args:
            group_fn_display_name (str): group function's display name
        """
        if group_fn_display_name != self._ui.group_fn_combo_box.currentText():
            self._ui.group_fn_combo_box.currentTextChanged.disconnect(self._change_root_mapping_group_fn)
            self._ui.group_fn_combo_box.setCurrentText(group_fn_display_name)
            self._ui.group_fn_combo_box.currentTextChanged.connect(self._change_root_mapping_group_fn)

    def _change_highlight_dimension(self, dimension):
        """Pushes a command to change highlight dimension to undo stack.

        Args:
            dimension (int): highlight dimension
        """
        if self._ui.item_type_combo_box.currentText() != "Entity class with dimension parameter":
            return
        index = self._sort_mappings_table_model.mapToSource(self._ui.mappings_table.currentIndex())
        if not index.isValid():
            return
        old = index.data(MappingsTableModel.HIGHLIGHT_POSITION_ROLE)
        new = dimension - 1
        if old == new:
            return
        self._undo_stack.push(SetHighlightDimension(index, old, new))

    def _set_highlight_dimension_silently(self, highlight_dimension):
        """Sets highlight dimension without emitting signals.

        Args:
            highlight_dimension (int, optional): highlight dimension
        """
        self._ui.highlight_dimension_spin_box.valueChanged.disconnect(self._change_highlight_dimension)
        if highlight_dimension is not None:
            self._ui.highlight_dimension_spin_box.setValue(highlight_dimension + 1)
            self._ui.highlight_dimension_spin_box.setEnabled(True)
        else:
            self._ui.highlight_dimension_spin_box.setValue(1)
            self._ui.highlight_dimension_spin_box.setEnabled(False)
        self._ui.highlight_dimension_spin_box.valueChanged.connect(self._change_highlight_dimension)

    def _change_entity_dimensions(self, dimensions):
        """
        Pushes a command to undo stack.

        Args:
            dimensions (int): dimensions
        """
        index = self._sort_mappings_table_model.mapToSource(self._ui.mappings_table.currentIndex())
        mapping = self._mappings_table_model.data(index, MappingsTableModel.MAPPING_ROOT_ROLE)
        modified = deepcopy(mapping)
        set_entity_dimensions(modified, dimensions)
        current_mapping_type = self._ui.item_type_combo_box.currentText()
        self._undo_stack.beginMacro("change entity dimensions")
        self._undo_stack.push(SetMapping(self._ui.mappings_table.currentIndex(), modified))
        if current_mapping_type == "Entity class with dimension parameter":
            highlight_dimension = self._mappings_table_model.data(index, MappingsTableModel.HIGHLIGHT_POSITION_ROLE)
            if highlight_dimension is None:
                self._undo_stack.push(SetHighlightDimension(index, highlight_dimension, 0))
            elif highlight_dimension >= dimensions:
                new_highlight_dimension = dimensions - 1 if dimensions > 0 else None
                self._undo_stack.push(SetHighlightDimension(index, highlight_dimension, new_highlight_dimension))
        self._undo_stack.endMacro()

    def _set_entity_dimensions_silently(self, dimensions):
        """
        Sets entity dimensions spin box without emitting signals.

        Args:
            dimensions (int): dimensions
        """
        self._ui.entity_dimensions_spin_box.valueChanged.disconnect(self._change_entity_dimensions)
        self._ui.entity_dimensions_spin_box.setValue(dimensions)
        self._ui.entity_dimensions_spin_box.valueChanged.connect(self._change_entity_dimensions)
        self._ui.highlight_dimension_spin_box.setMaximum(max(dimensions, 1))

    def _enable_mapping_specification_editing(self):
        """Enables and disables mapping specification editing controls."""
        have_mappings = self._mappings_table_model.rowCount() > 0
        self._ui.mapping_options_contents.setEnabled(have_mappings)
        self._ui.mapping_table_view.setEnabled(have_mappings)
        self._ui.remove_mapping_button.setEnabled(have_mappings)

    def _enable_entity_controls(self):
        """Enables and disables controls related to entity export."""
        current_mapping_type = self._ui.item_type_combo_box.currentText()
        self._ui.entity_dimensions_spin_box.setEnabled(current_mapping_type.startswith("Entity class"))
        self._ui.highlight_dimension_spin_box.setEnabled(
            current_mapping_type == "Entity class with dimension parameter"
        )

    def _enable_parameter_controls(self):
        """Enables and disables controls related to entity export."""
        mapping_type = self._ui.item_type_combo_box.currentText()
        types_with_parameters = {"Entity class", "Entity class with dimension parameter"}
        if mapping_type in types_with_parameters:
            self._ui.parameter_type_combo_box.setEnabled(True)
            self._ui.parameter_type_combo_box.currentTextChanged.disconnect(self._change_mapping_type)
            model = self._ui.parameter_type_combo_box.model()
            default_value_item = model.item(1)
            default_value_item.setFlags(default_value_item.flags() | Qt.ItemIsEnabled)
            no_value_item = model.item(2)
            if mapping_type == "Entity class with dimension parameter":
                no_value_item.setFlags(default_value_item.flags() & ~Qt.ItemIsEnabled)
            else:
                no_value_item.setFlags(default_value_item.flags() | Qt.ItemIsEnabled)
            self._ui.parameter_type_combo_box.currentTextChanged.connect(self._change_mapping_type)
            self._ui.parameter_dimensions_spin_box.setEnabled(self._ui.parameter_type_combo_box.currentText() != "None")
        else:
            self._ui.parameter_type_combo_box.setEnabled(False)
            self._ui.parameter_dimensions_spin_box.setEnabled(False)

    @Slot()
    def _compact_mapping(self):
        """Pushes a CompactMapping command to the undo stack."""
        if not self._mapping_editor_model.can_compact():
            return
        row = self._ui.mappings_table.selectionModel().currentIndex().row()
        mapping_name = self._sort_mappings_table_model.mapToSource(self._sort_mappings_table_model.index(row, 0)).data()
        self._undo_stack.push(CompactMapping(self._mapping_editor_model, mapping_name))

    def tear_down(self):
        if not super().tear_down():
            return False
        self._preview_updater.tear_down()
        return True


def _new_mapping_specification(mapping_type):
    """
    Creates a new export mapping.

    Args:
        mapping_type (MappingType): mapping's type

    Returns:
        MappingSpecification: an export mapping specification
    """
    if mapping_type == MappingType.entities:
        return MappingSpecification(mapping_type, True, True, NoGroup.NAME, False, entity_export(0, 1))
    if mapping_type == MappingType.entity_groups:
        return MappingSpecification(mapping_type, True, True, NoGroup.NAME, False, entity_group_export(0, 1, 2))
    if mapping_type == MappingType.entity_parameter_default_values:
        return MappingSpecification(
            mapping_type,
            True,
            True,
            NoGroup.NAME,
            False,
            entity_parameter_default_value_export(0, 1, Position.hidden, 2, None, None),
        )
    if mapping_type == MappingType.entity_parameter_values:
        return MappingSpecification(
            mapping_type,
            True,
            True,
            NoGroup.NAME,
            False,
            entity_parameter_value_export(0, 2, Position.hidden, 1, None, None, 3, Position.hidden, 4, None, None),
        )
    if mapping_type == MappingType.parameter_value_lists:
        return MappingSpecification(mapping_type, True, True, NoGroup.NAME, False, parameter_value_list_export(0, 1))
    if mapping_type == MappingType.entity_dimension_parameter_default_values:
        return MappingSpecification(
            mapping_type,
            True,
            True,
            NoGroup.NAME,
            False,
            entity_dimension_parameter_default_value_export(0, 1, None, Position.hidden, 2, None, None, 0),
        )
    if mapping_type == MappingType.entity_dimension_parameter_values:
        return MappingSpecification(
            mapping_type,
            True,
            True,
            NoGroup.NAME,
            False,
            entity_dimension_parameter_value_export(
                0, 3, Position.hidden, Position.hidden, [1], [2], 4, Position.hidden, 5, None, None, 0
            ),
        )
    if mapping_type == MappingType.alternatives:
        return MappingSpecification(mapping_type, True, True, NoGroup.NAME, False, alternative_export(0))
    if mapping_type == MappingType.scenario_alternatives:
        return MappingSpecification(mapping_type, True, True, NoGroup.NAME, False, scenario_alternative_export(0, 1))
    if mapping_type == MappingType.scenarios:
        return MappingSpecification(mapping_type, True, True, NoGroup.NAME, False, scenario_export(0))
    raise NotImplementedError()


def _add_fixed_table_name(mapping_root):
    """
    Adds a fixed table name to given mapping.

    Args:
        mapping_root (Mapping): a root mapping

    Returns:
        Mapping: new mapping hierarchy
    """
    for mapping in mapping_root.flatten():
        if mapping.position == Position.table_name:
            mapping.position = Position.hidden
    new_root = FixedValueMapping(Position.table_name, "table")
    new_root.child = mapping_root
    return new_root


def _remove_fixed_table_name(mapping_root):
    """
    Removes fixed table name from mapping.

    Args:
        mapping_root (Mapping): a root mapping

    Returns:
        Mapping: new mapping hierarchy
    """
    return unflatten(mapping_root.flatten()[1:])



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains mapping position editor delegate."""
from PySide6.QtCore import Property, Slot
from PySide6.QtWidgets import QComboBox, QLineEdit, QStyledItemDelegate
from ..mvcmodels.mapping_editor_table_model import POSITION_DISPLAY_TEXT

_positions = list(POSITION_DISPLAY_TEXT.values())


class PositionEditDelegate(QStyledItemDelegate):
    """Custom delegate for editing positions."""

    def createEditor(self, parent, option, index):
        return _PositionEdit(parent)


class _PositionEdit(QComboBox):
    def __init__(self, parent):
        """
        Args:
            parent (QWidget): parent widget
        """
        super().__init__(parent)
        self.addItems(_positions)
        self.setEditable(True)
        self.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)
        self.currentTextChanged.connect(self._insert)
        self.setLineEdit(_SelectingLineEdit())

    @Slot(str)
    def _insert(self, text):
        """
        Inserts new text item at the top of the box or replaces the current item.

        Args:
            text (str): text to insert
        """
        if text not in _positions:
            if self.count() == len(_positions):
                self.insertItem(0, text)
            else:
                self.setItemText(0, text)

    def set_position(self, position):
        """
        Sets the combo boxes value.

        Args:
            position (Position or str): position
        """
        if not position.isdigit():
            self.setCurrentText(position)
        else:
            self._insert(position)
            self.setCurrentIndex(0)

    def position(self):
        """Gets the position.

        Returns:
            str: position
        """
        return self.currentText()

    regexp = Property(str, position, set_position, user=True)
    """Property used to communicate with the editor delegate."""


class _SelectingLineEdit(QLineEdit):
    """Line editor that selects all text when focussed."""

    def focusInEvent(self, e):
        self.selectAll()
        super().focusInEvent(e)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Module for MergerIcon class."""
import random
from PySide6.QtCore import QPointF, QTimeLine, Slot
from spinetoolbox.project_item_icon import ProjectItemIcon
from ..animations import AnimationSignaller


class MergerIcon(ProjectItemIcon):
    _SHAKE_FACTOR = 0.05

    def __init__(self, toolbox, icon, icon_color):
        """Data Store icon for the Design View.

        Args:
            toolbox (ToolBoxUI): QMainWindow instance
            icon (str): icon resource path
            icon_color (QColor): Icon's color
        """
        super().__init__(toolbox, icon, icon_color)
        self._time_line = QTimeLine()
        self._time_line.setLoopCount(0)  # loop forever
        self._time_line.setFrameRange(0, 10)
        self._time_line.setDirection(QTimeLine.Direction.Backward)
        self._time_line.valueChanged.connect(self._handle_time_line_value_changed)
        self._time_line.stateChanged.connect(self._handle_time_line_state_changed)
        self._svg_item_pos = self.svg_item.pos()
        self.animation_signaller = AnimationSignaller()
        self.animation_signaller.animation_started.connect(self.start_animation)
        self.animation_signaller.animation_stopped.connect(self.stop_animation)

    @Slot(float)
    def _handle_time_line_value_changed(self, value):
        rect = self.svg_item.sceneBoundingRect()
        width = rect.width()
        height = rect.height()
        x = random.uniform(-self._SHAKE_FACTOR, self._SHAKE_FACTOR) * width
        y = random.uniform(-self._SHAKE_FACTOR, self._SHAKE_FACTOR) * height
        self.svg_item.setPos(self._svg_item_pos + QPointF(x, y))

    @Slot(QTimeLine.State)
    def _handle_time_line_state_changed(self, new_state):
        if new_state == QTimeLine.State.NotRunning:
            self.svg_item.setPos(self._svg_item_pos)

    @Slot()
    def start_animation(self):
        """Start the animation that plays when the Merger associated to this GraphicsItem is running."""
        if self._time_line.state() == QTimeLine.State.Running:
            return
        self._time_line.start()

    @Slot()
    def stop_animation(self):
        """Stop animation"""
        if self._time_line.state() != QTimeLine.State.Running:
            return
        self._time_line.stop()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for ``tool_specifications`` module."""
import os
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from spine_engine.project_item.connection import Jump
from spine_engine.spine_engine import SpineEngine
from spine_items.tool.tool_specifications import (
    ExecutableTool,
    GAMSTool,
    JuliaTool,
    PythonTool,
    ToolSpecification,
    make_specification,
)
from tests.mock_helpers import MockQSettings


class TestToolSpecification(unittest.TestCase):
    def setUp(self) -> None:
        self.qsettings = MockQSettings()
        self.logger = mock.MagicMock()
        self.test_dict = {
            "name": "specification name",
            "tooltype": "Python",
            "includes": ["/included_file_1.bat", "/aa/included_file_2.std"],
            "description": "Test tool specification",
            "inputfiles": ["a_input.jpg", "b_input.png"],
            "inputfiles_opt": ["*.dat", "*.xlsx", "a.csv", "z.zip"],
            "outputfiles": ["*.atk", "*.zip"],
            "cmdline_args": ["99", "10"],
            "includes_main_path": "../tool",
        }

    def test_to_dict_sorts_input_and_output_files(self):
        specification = ToolSpecification(
            "specification name",
            "Python",
            "/path/to/tool",
            ["/included_file_1.bat", "/aa/included_file_2.std"],
            self.qsettings,
            self.logger,
            "Test tool specification",
            ["b_input.png", "a_input.jpg"],
            ["*.dat", "a.csv", "z.zip", "*.xlsx"],
            ["*.zip", "*.atk"],
            ["99", "10"],
        )
        specification.definition_file_path = "/path/to/specification/file.json"
        specification_dict = specification.to_dict()
        self.assertEqual(specification_dict, self.test_dict)

    def test_make_specification(self):
        self.test_dict["definition_file_path"] = "/path/to/specification/file.json"
        spec = make_specification(self.test_dict, self.qsettings, self.logger)  # Make PythonTool
        self.assertIsInstance(spec, PythonTool)
        spec.init_execution_settings()
        self.assertIsNotNone(spec.execution_settings)
        self.assertTrue(len(spec.execution_settings.keys()), 4)
        spec.to_dict()
        # Convert to Julia spec
        self.test_dict["tooltype"] = "Julia"
        spec = make_specification(self.test_dict, self.qsettings, self.logger)  # Make JuliaTool
        self.assertIsInstance(spec, JuliaTool)
        spec.init_execution_settings()
        self.assertIsNotNone(spec.execution_settings)
        self.assertTrue(len(spec.execution_settings.keys()), 5)
        spec.to_dict()
        # Convert to GAMS spec
        self.test_dict["tooltype"] = "GAMS"
        spec = make_specification(self.test_dict, self.qsettings, self.logger)  # Make GAMSTool
        self.assertIsInstance(spec, GAMSTool)
        spec.to_dict()
        # Convert to Executable spec
        self.test_dict["tooltype"] = "Executable"
        spec = make_specification(self.test_dict, self.qsettings, self.logger)  # Make ExecutableTool
        self.assertIsInstance(spec, ExecutableTool)
        spec.init_execution_settings()
        self.assertIsNotNone(spec.execution_settings)
        self.assertTrue(len(spec.execution_settings.keys()), 2)
        spec.to_dict()

    def test_clone(self):
        self.test_dict["definition_file_path"] = "/path/to/specification/file.json"
        spec = make_specification(self.test_dict, self.qsettings, self.logger)
        cloned_spec = spec.clone()
        self.assertFalse(spec.is_equivalent(cloned_spec))  # False because 'path' is different. Bug?

    def test_tool_specification_as_jump_condition(self):
        condition = {"type": "tool-specification", "specification": "loop_twice"}
        jump = Jump("source", "bottom", "destination", "top", condition)
        jump.make_logger(mock.Mock())
        with TemporaryDirectory() as temp_dir:
            main_file = "script.py"
            main_file_path = os.path.join(temp_dir, main_file)
            with open(main_file_path, "w+") as program_file:
                program_file.writelines(
                    ["import sys\n", "counter = int(sys.argv[1])\n", "exit(0 if counter == 23 else 1)\n"]
                )
            spec_dict = {
                "name": "loop_twice",
                "tooltype": "python",
                "includes_main_path": temp_dir,
                "includes": [main_file],
                "definition_file_path": "path/to/specification_file.json",
            }
            engine = SpineEngine(project_dir=temp_dir, specifications={"Tool": [spec_dict]}, connections=[])
            jump.set_engine(engine)
            self.assertTrue(jump.is_condition_true(23))


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains Exporter's undo commands."""
from copy import copy, deepcopy
from enum import IntEnum, unique
from PySide6.QtCore import QModelIndex, Qt  # pylint: disable=unused-import
from PySide6.QtGui import QUndoCommand
from spinetoolbox.helpers import SealCommand
from spinetoolbox.project_commands import SpineToolboxCommand
from .mvcmodels.mappings_table_model import MappingsTableModel


@unique
class CommandId(IntEnum):
    CHANGE_POSITION = 1
    CHANGE_FIXED_TABLE_NAME = 2
    CHANGE_OUT_LABEL = 3


class UpdateOutLabel(SpineToolboxCommand):
    """Command to update exporter's output label."""

    def __init__(self, exporter_name, out_label, in_label, previous_label, project):
        """
        Args:
            exporter_name (str): exporter's name
            out_label (str): new output resource label
            in_label (str): associated input resource label
            previous_label (str): previous output resource label
            project (SpineToolboxProject): project
        """
        super().__init__()
        self._exporter_name = exporter_name
        self._out_label = out_label
        self._previous_out_label = previous_label
        self._in_label = in_label
        self._project = project
        self.setText(f"change output label in {exporter_name}")
        self._sealed = False

    def id(self):
        return 1

    def redo(self):
        exporter = self._project.get_item(self._exporter_name)
        exporter.set_out_label(self._out_label, self._in_label)

    def undo(self):
        exporter = self._project.get_item(self._exporter_name)
        exporter.set_out_label(self._previous_out_label, self._in_label)

    def mergeWith(self, other):
        if not self._sealed:
            if (
                isinstance(other, UpdateOutLabel)
                and self._exporter_name == other._exporter_name
                and self._in_label == other._in_label
            ):
                if self._previous_out_label != other._out_label:
                    self._out_label = other._out_label
                else:
                    self.setObsolete(True)
                return True
            if isinstance(other, SealCommand):
                self._sealed = True
                return True
        return False


class UpdateOutUrl(SpineToolboxCommand):
    """Command to update exporter's output URL."""

    def __init__(self, exporter_name, in_label, url, previous_url, project):
        """
        Args:
            exporter_name (str): exporter's name
            in_label (str): input resource label
            url (dict, optional): new URL dict
            previous_url (dict, optional): previous URL dict
            project (SpineToolboxProject): project
        """
        super().__init__()
        self._exporter_name = exporter_name
        self._in_label = in_label
        self._url = copy(url)
        self._previous_url = copy(previous_url)
        self._project = project
        self.setText(f"change output URL in {exporter_name}")

    def redo(self):
        exporter = self._project.get_item(self._exporter_name)
        exporter.set_out_url(self._in_label, copy(self._url))

    def undo(self):
        exporter = self._project.get_item(self._exporter_name)
        exporter.set_out_url(self._in_label, copy(self._previous_url))


class NewMapping(QUndoCommand):
    def __init__(self, mappings_table_model, mapping_specification, name=""):
        """
        Args:
            mappings_table_model (MappingsTableModel): mapping table model
            mapping_specification (MappingSpecification): new mapping specification
            name (str): specification's name
        """
        super().__init__("add mapping")
        self._mappings_table_model = mappings_table_model
        self._mapping_specification = mapping_specification
        self._mapping_name = name

    def redo(self):
        row = self._mappings_table_model.extend(deepcopy(self._mapping_specification), self._mapping_name)
        self._mapping_name = self._mappings_table_model.index(row, 0).data()

    def undo(self):
        self._mappings_table_model.remove_mapping(self._mapping_name)


class RenameMapping(QUndoCommand):
    """A command to change the name of a mapping."""

    def __init__(self, row, mapping_list_model, name):
        """
        Args:
            row (int): row index
            mapping_list_model (MappingListModel): model holding the mapping names
            name (str): new name
        """
        text = "rename mapping"
        super().__init__(text)
        self._row = row
        self._model = mapping_list_model
        self._name = name
        self._previous_name = self._model.index(self._row, 0).data()

    def redo(self):
        """Renames the mapping."""
        self._model.rename_mapping(self._row, self._name)

    def undo(self):
        """Reverts renaming of the mapping."""
        self._model.rename_mapping(self._row, self._previous_name)


class RemoveMapping(QUndoCommand):
    def __init__(self, row, mappings_table_model):
        """
        Args:
            row (int): row index of mapping's name
            mappings_table_model (MappingsTableModel): mapping list model
        """
        super().__init__("remove mapping")
        self._mappings_table_model = mappings_table_model
        self._row = row
        index = self._mappings_table_model.index(row, 0)
        self._name = index.data()
        self._mapping_specification = index.data(MappingsTableModel.MAPPING_SPECIFICATION_ROLE)

    def redo(self):
        self._mappings_table_model.remove_mapping(self._name)

    def undo(self):
        self._mappings_table_model.insert_mapping(self._row, self._name, self._mapping_specification)


class SetMappingEnabled(QUndoCommand):
    def __init__(self, row, mappings_table_model):
        """
        Args:
            row (int): row index of mapping's name
            mappings_table_model (MappingsTableModel): mapping list model
        """
        self._mappings_table_model = mappings_table_model
        name = self._mappings_table_model.index(row, 0).data()
        self._row = row
        self._previously_enabled = (
            self._mappings_table_model.index(self._row, 0).data(Qt.ItemDataRole.CheckStateRole).value
            == Qt.CheckState.Checked.value
        )
        super().__init__(("disable" if self._previously_enabled else "enable") + f" '{name}'")

    def redo(self):
        self._mappings_table_model.set_mapping_enabled(self._row, not self._previously_enabled)

    def undo(self):
        self._mappings_table_model.set_mapping_enabled(self._row, self._previously_enabled)


class EnableAllMappings(QUndoCommand):
    def __init__(self, mappings_table_model):
        """
        Args:
            mappings_table_model (MappingsTableModel): mapping list model
        """
        super().__init__("enable all mappings")
        self._mappings_table_model = mappings_table_model
        self._previously_enabled = self._mappings_table_model.enabled_mapping_rows()

    def redo(self):
        self._mappings_table_model.set_all_enabled(True)

    def undo(self):
        self._mappings_table_model.enable_mapping_rows(self._previously_enabled)


class DisableAllMappings(QUndoCommand):
    def __init__(self, mappings_table_model):
        """
        Args:
            mappings_table_model (MappingsTableModel): mapping list model
        """
        super().__init__("disable all mappings")
        self._mappings_table_model = mappings_table_model

    def redo(self):
        self._mappings_table_model.set_all_enabled(False)

    def undo(self):
        self._mappings_table_model.set_all_enabled(True)


class ChangeWriteOrder(QUndoCommand):
    def __init__(self, row, earlier, mappings_table_model):
        """
        Args:
            row (int): row index of mapping's name
            earlier (bool): True to write mapping earlier, False to write later
            mappings_table_model (MappingsTableModel): mappings table model
        """
        super().__init__("change writing order")
        self._row = row
        self._earlier = earlier
        self._mappings_table_model = mappings_table_model

    def redo(self):
        self._mappings_table_model.reorder_writing(self._row, self._earlier)

    def undo(self):
        self._mappings_table_model.reorder_writing(self._row - 1 if self._earlier else self._row + 1, not self._earlier)


class SetMapping(QUndoCommand):
    def __init__(self, index, mapping):
        """
        Args:
            index (QModelIndex): mapping's row index
            mapping (Mapping): mapping root
        """
        super().__init__("mapping change")
        self._index = index
        self._mapping = mapping
        self._previous_mapping = self._index.data(MappingsTableModel.MAPPING_ROOT_ROLE)

    def redo(self):
        self._index.model().setData(self._index, deepcopy(self._mapping), MappingsTableModel.MAPPING_ROOT_ROLE)

    def undo(self):
        self._index.model().setData(self._index, deepcopy(self._previous_mapping), MappingsTableModel.MAPPING_ROOT_ROLE)


class SetMappingType(QUndoCommand):
    def __init__(self, index, type_):
        """
        Args:
            index (QModelIndex): mapping's row index
            type_ (MappingType): mapping type
        """
        super().__init__("mapping type change")
        self._index = index
        self._type = type_
        self._previous_type = self._index.data(MappingsTableModel.MAPPING_TYPE_ROLE)

    def redo(self):
        self._index.model().setData(self._index, self._type, MappingsTableModel.MAPPING_TYPE_ROLE)

    def undo(self):
        self._index.model().setData(self._index, self._previous_type, MappingsTableModel.MAPPING_TYPE_ROLE)


class SetAlwaysExportHeader(QUndoCommand):
    def __init__(self, index, always_export_header):
        """
        Args:
            index (QModelIndex): mapping's row index
            always_export_header (bool): always export header flag
        """
        super().__init__(("check" if always_export_header else "uncheck") + " always export header checkbox")
        self._index = index
        self._flag = always_export_header

    def redo(self):
        self._index.model().setData(self._index, self._flag, MappingsTableModel.ALWAYS_EXPORT_HEADER_ROLE)

    def undo(self):
        self._index.model().setData(self._index, not self._flag, MappingsTableModel.ALWAYS_EXPORT_HEADER_ROLE)


class SetUseFixedTableNameFlag(QUndoCommand):
    def __init__(self, index, flag):
        """
        Args:
            index (QModelIndex): mapping's row index
            flag (bool): use fixed table name flag
        """
        super().__init__(("check" if flag else "uncheck") + " fixed table name checkbox")
        self._index = index
        self._flag = flag

    def redo(self):
        self._index.model().setData(self._index, self._flag, MappingsTableModel.USE_FIXED_TABLE_NAME_FLAG_ROLE)

    def undo(self):
        self._index.model().setData(self._index, not self._flag, MappingsTableModel.USE_FIXED_TABLE_NAME_FLAG_ROLE)


class SetFixedTableName(QUndoCommand):
    ID = CommandId.CHANGE_FIXED_TABLE_NAME.value

    def __init__(self, index, old_name, new_name):
        """
        Args:
            index (QModelIndex): mapping's row index
            old_name (str): old fixed table name
            new_name (str): new fixed table name
        """
        super().__init__("change fixed table name")
        self._index = index
        self._old_name = old_name
        self._new_name = new_name
        self._sealed = False

    def redo(self):
        self._index.model().setData(self._index, self._new_name, MappingsTableModel.FIXED_TABLE_NAME_ROLE)

    def undo(self):
        self._index.model().setData(self._index, self._old_name, MappingsTableModel.FIXED_TABLE_NAME_ROLE)

    def id(self):
        return self.ID

    def mergeWith(self, other):
        if not self._sealed:
            if isinstance(other, SetFixedTableName) and self.id() == other.id() and self._index == other._index:
                if self._old_name != other._new_name:
                    self._new_name = other._new_name
                else:
                    self.setObsolete(True)
                return True
            if isinstance(other, SealCommand):
                self._sealed = True
                return True
        return False


class SetGroupFunction(QUndoCommand):
    def __init__(self, index, old_function, new_function):
        """
        Args
            index (QModelIndex): mapping's row index
            old_function (str): old group function's name
            new_function (str): new group function's name
        """
        super().__init__("change group function")
        self._index = index
        self._old_function = old_function
        self._new_function = new_function

    def redo(self):
        self._index.model().setData(self._index, self._new_function, MappingsTableModel.GROUP_FN_ROLE)

    def undo(self):
        self._index.model().setData(self._index, self._old_function, MappingsTableModel.GROUP_FN_ROLE)


class SetHighlightDimension(QUndoCommand):
    def __init__(self, index, old_dimension, new_dimension):
        """
        Args
            index (QModelIndex): mapping's row index
            old_dimension (int): old highlight dimension
            new_dimension (int): new highlight dimension
        """
        super().__init__("change selected relationship dimension")
        self._index = index
        self._old_dimension = old_dimension
        self._new_dimension = new_dimension

    def redo(self):
        self._index.model().setData(self._index, self._new_dimension, MappingsTableModel.HIGHLIGHT_POSITION_ROLE)

    def undo(self):
        self._index.model().setData(self._index, self._old_dimension, MappingsTableModel.HIGHLIGHT_POSITION_ROLE)


class SetMappingPositions(QUndoCommand):
    def __init__(self, model, mapping_name, positions, previous_positions):
        """
        Args:
            model (MappingEditorTableModel): editor model
            mapping_name (str): mapping's name
            positions (list of Position): new positions
            previous_positions (list of Position): previous positions
        """
        super().__init__("change mapping item's position")
        self._mapping_editor_table_model = model
        self._mapping_name = mapping_name
        self._positions = positions
        self._previous_positions = previous_positions

    @property
    def mapping_editor_table_model(self):
        return self._mapping_editor_table_model

    @property
    def mapping_name(self):
        return self._mapping_name

    @property
    def positions(self):
        return self._positions

    @property
    def previous_positions(self):
        return self._previous_positions

    def id(self):
        return int(CommandId.CHANGE_POSITION)

    def redo(self):
        self._mapping_editor_table_model.set_positions(self._positions, self._mapping_name)

    def undo(self):
        self._mapping_editor_table_model.set_positions(self._previous_positions, self._mapping_name)


class ClearFixedTableName(QUndoCommand):
    def __init__(self, editor, is_fixed_table_checked, previous_fixed_table_name, mapping_index, mapping_root):
        """
        Args:
            editor (SpecificationEditor): specification editor window
            is_fixed_table_checked (bool): state of the fixed table name checkbox before clearing
            previous_fixed_table_name (str): fixed table name before clearing
            mapping_index (QModelIndex): mapping's row index
            mapping_root (Mapping): mapping root
        """
        super().__init__("change mapping item's position")
        self._editor = editor
        self._is_fixed_table_checked = is_fixed_table_checked
        self._previous_fixed_table_name = previous_fixed_table_name
        self._mapping_editor_table_model = None
        self._mapping_name = None
        self._positions = None
        self._previous_positions = None
        self._mapping_index = mapping_index
        self._mapping_root = mapping_root
        self._previous_mapping_root = self._mapping_index.data(MappingsTableModel.MAPPING_ROOT_ROLE)

    def id(self):
        return int(CommandId.CHANGE_POSITION)

    def mergeWith(self, other):
        if self._mapping_editor_table_model is not None:
            return False
        self._mapping_editor_table_model = other.mapping_editor_table_model
        self._mapping_name = other.mapping_name
        self._positions = other.positions
        self._previous_positions = other.previous_positions
        return True

    def redo(self):
        self._editor.clear_fixed_table_name()
        self._mapping_index.model().setData(
            self._mapping_index, deepcopy(self._mapping_root), MappingsTableModel.MAPPING_ROOT_ROLE
        )

    def undo(self):
        self._mapping_index.model().setData(
            self._mapping_index, deepcopy(self._previous_mapping_root), MappingsTableModel.MAPPING_ROOT_ROLE
        )
        self._editor.enable_fixed_table_name(self._is_fixed_table_checked, self._previous_fixed_table_name)
        if self._mapping_editor_table_model is not None:
            self._mapping_editor_table_model.set_positions(self._previous_positions, self._mapping_name)


class SetMappingProperty(QUndoCommand):
    def __init__(self, command_name, setter, mapping_name, row, value, previous_value):
        """
        Sets either ``header`` or ``filter_re``.

        Args:
            command_name (str)
            setter (Callable): setter function
            mapping_name (str): mapping's name
            row (int): model row
            value (str): mapping's new value for the property
            previous_value (str): previous value
        """
        super().__init__(command_name)
        self._setter = setter
        self._mapping_name = mapping_name
        self._row = row
        self._value = value
        self._previous_value = previous_value

    def redo(self):
        self._setter(self._value, self._row, self._mapping_name)

    def undo(self):
        self._setter(self._previous_value, self._row, self._mapping_name)


class SetMappingNullable(QUndoCommand):
    def __init__(self, model, mapping_name, row, nullable):
        """
        Args:
            model (MappingEditorTableModel): editor model
            mapping_name (str): mapping's name
            row (int): row index in model
            nullable (bool): True to set, False to unset
        """
        super().__init__(("set " if nullable else "unset ") + " mapping item nullable")
        self._model = model
        self._mapping_name = mapping_name
        self._row = row
        self._nullable = nullable

    def redo(self):
        self._model.set_nullable(self._nullable, self._row, self._mapping_name)

    def undo(self):
        self._model.set_nullable(not self._nullable, self._row, self._mapping_name)


class CompactMapping(QUndoCommand):
    def __init__(self, model, mapping_name):
        """
        Args:
            model (MappingEditorTableModel): editor model
            mapping_name (str): mapping's name
        """
        super().__init__("compact mapping")
        self._model = model
        self._mapping_name = mapping_name
        self._previous_positions = self._model.positions()

    def redo(self):
        self._model.compact()

    def undo(self):
        self._model.set_positions(self._previous_positions, self._mapping_name)


class SetExportFormat(QUndoCommand):
    def __init__(self, editor, export_format, previous_format):
        """
        Args:
            editor (SpecificationEditor): specification editor window
            export_format (OutputFormat): new format
            previous_format (OutputFormat): previous format
        """
        super().__init__("change export format")
        self._specification_editor = editor
        self._format = export_format
        self._previous_format = previous_format

    def redo(self):
        self._specification_editor.set_export_format_silently(self._format)

    def undo(self):
        self._specification_editor.set_export_format_silently(self._previous_format)


class UpdateOutputTimeStampsFlag(SpineToolboxCommand):
    """Command to set exporter's output directory time stamps flag."""

    def __init__(self, exporter_name, value, project):
        """
        Args:
            exporter_name (str): exporter's name
            value (bool): flag's new value
            project (SpineToolboxProject): project
        """
        super().__init__()
        self.setText(f"toggle output time stamps setting of {exporter_name}")
        self._exporter_name = exporter_name
        self._value = value
        self._project = project

    def redo(self):
        exporter = self._project.get_item(self._exporter_name)
        exporter.set_output_time_stamps_flag(self._value)

    def undo(self):
        exporter = self._project.get_item(self._exporter_name)
        exporter.set_output_time_stamps_flag(not self._value)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Module for data connection icon class."""
import os
from PySide6.QtCore import QObject, Qt, QTimer, Signal
from PySide6.QtWidgets import QGraphicsItem
from spinetoolbox.project_item_icon import ProjectItemIcon


class DataConnectionIcon(ProjectItemIcon):
    class _SignalHolder(QObject):
        files_dropped_on_icon = Signal(QGraphicsItem, list)
        """A signal that it triggered when files are dragged and dropped on the item."""

    def __init__(self, toolbox, icon, icon_color):
        """Data Connection icon for the Design View.

        Args:
            toolbox (ToolboxUI): main window instance
            icon (str): icon resource path
            icon_color (QColor): Icon's color
        """
        super().__init__(toolbox, icon, icon_color)
        self.setAcceptDrops(True)
        self._drag_over = False
        self._signal_holder = DataConnectionIcon._SignalHolder()
        self.files_dropped_on_icon = self._signal_holder.files_dropped_on_icon

    def dragEnterEvent(self, event):
        """Drag and drop action enters.
        Accept file drops from the filesystem.

        Args:
            event (QGraphicsSceneDragDropEvent): Event
        """
        urls = event.mimeData().urls()
        for url in urls:
            if not url.isLocalFile():
                event.ignore()
                return
            if not os.path.isfile(url.toLocalFile()):
                event.ignore()
                return
        event.accept()
        event.setDropAction(Qt.CopyAction)
        if self._drag_over:
            return
        self._drag_over = True
        QTimer.singleShot(100, self.select_on_drag_over)

    def dragLeaveEvent(self, event):
        """Drag and drop action leaves.

        Args:
            event (QGraphicsSceneDragDropEvent): Event
        """
        event.accept()
        self._drag_over = False

    def dragMoveEvent(self, event):
        """Accept event."""
        event.accept()

    def dropEvent(self, event):
        """Emit files_dropped_on_dc signal from scene,
        with this instance, and a list of files for each dropped url."""
        self.files_dropped_on_icon.emit(self, [url.toLocalFile() for url in event.mimeData().urls()])

    def select_on_drag_over(self):
        """Called when the timer started in drag_enter_event is elapsed.
        Select this item if the drag action is still over it.
        """
        if not self._drag_over:
            return
        self._drag_over = False
        self._toolbox.ui.graphicsView.scene().clearSelection()
        self.setSelected(True)



######################################################################################################################
# Copyright (C) 2017-2023 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Exporter's ``FullUrlListModel``."""
from PySide6.QtCore import QAbstractListModel, QModelIndex, Qt


class FullUrlListModel(QAbstractListModel):
    def __init__(self, parent=None):
        """
        Args:
            parent (QObject, optional): model's parent
        """
        super().__init__(parent)
        self._urls = []

    def append(self, url):
        """
        Appends a URL to the model.

        Args:
            url (str): URL to append
        """
        n = len(self._urls)
        self.beginInsertRows(QModelIndex(), n, n)
        self._urls.append(url)
        self.endInsertRows()

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None
        if role == Qt.ItemDataRole.DisplayRole:
            return self._urls[index.row()]
        return None

    def rowCount(self, parent=QModelIndex()):
        return len(self._urls)

    def set_urls(self, urls):
        """
        Sets model's URLs.

        Args:
            urls (Iterable of str): URLs
        """
        self.beginResetModel()
        self._urls = list(urls)
        self.endResetModel()

    def update_url(self, old, new):
        """
        Updates a database URL.

        Args:
            old (str): old URL
            new (str): new URL
        """
        for row, url in enumerate(self._urls):
            if old == url:
                self._urls[row] = new
                index = self.index(row, 0)
                self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole])
                return



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

""" Unit tests for the ``utils`` module. """
import sys
import unittest
from unittest import mock
from spine_items.utils import (
    convert_to_sqlalchemy_url,
    convert_url_to_safe_string,
    database_label,
    split_url_credentials,
    unsplit_url_credentials,
)


class TestDatabaseLabel(unittest.TestCase):
    def test_provider_name_is_appended_to_label(self):
        self.assertEqual(database_label("Item Name"), "db_url@Item Name")


class TestConvertToSqlAlchemyUrl(unittest.TestCase):
    def test_sqlite_url_conversion_succeeds(self):
        database_path = r"h:\files\database.sqlite" if sys.platform == "win32" else "/home/data/database.sqlite"
        url = {
            "dialect": "sqlite",
            "database": database_path,
        }
        sa_url = convert_to_sqlalchemy_url(url)
        self.assertEqual(sa_url.database, database_path)
        self.assertIsNone(sa_url.username)
        self.assertIsNone(sa_url.password)
        self.assertIsNone(sa_url.port)
        self.assertIsNone(sa_url.host)
        self.assertEqual(sa_url.drivername, r"sqlite")
        self.assertEqual(str(sa_url), r"sqlite:///" + database_path)

    def test_remote_url_conversion_succeeds(self):
        url = {
            "dialect": "mysql",
            "host": "example.com",
            "port": 5432,
            "database": "trade_nuc",
            "username": "superman",
            "password": "s3cr?t",
        }
        sa_url = convert_to_sqlalchemy_url(url)
        self.assertEqual(sa_url.database, "trade_nuc")
        self.assertEqual(sa_url.password, "s3cr?t")
        self.assertEqual(sa_url.username, "superman")
        self.assertEqual(sa_url.port, 5432)
        self.assertEqual(sa_url.host, "example.com")
        self.assertEqual(sa_url.drivername, r"mysql+pymysql")

    def test_remote_url_conversion_succeeds_with_schema(self):
        url = {
            "dialect": "mysql",
            "host": "example.com",
            "port": 5432,
            "database": "trade_nuc",
            "schema": "myschema",
            "username": "superman",
            "password": "s3cr?t",
        }
        sa_url = convert_to_sqlalchemy_url(url)
        self.assertEqual(sa_url.database, "trade_nuc")
        self.assertEqual(sa_url.password, "s3cr?t")
        self.assertEqual(sa_url.username, "superman")
        self.assertEqual(sa_url.port, 5432)
        self.assertEqual(sa_url.host, "example.com")
        self.assertEqual(sa_url.drivername, r"mysql+pymysql")

    def test_item_name_and_logger(self):
        url = {}
        logger = mock.MagicMock()
        self.assertIsNone(convert_to_sqlalchemy_url(url, "Item Name", logger))
        logger.msg_error.emit.assert_called_once_with(
            "No URL specified for <b>Item Name</b> selections. Please specify one and try again."
        )

    def test_missing_dialect_is_caught(self):
        url = {
            "dialect": "",
            "database": r"h:\files\database.sqlite",
        }
        logger = mock.MagicMock()
        self.assertIsNone(convert_to_sqlalchemy_url(url, "Item Name", logger))
        logger.msg_error.emit.assert_called_once_with(
            "Unable to generate URL from <b>Item Name</b> selections: missing dialect"
        )

    def test_missing_host_is_caught(self):
        url = {
            "dialect": "mysql",
            "host": "",
            "port": 5432,
            "database": "trade_nuc",
            "username": "superman",
            "password": "s3cr?t",
        }
        logger = mock.MagicMock()
        self.assertIsNone(convert_to_sqlalchemy_url(url, "Item Name", logger))
        logger.msg_error.emit.assert_called_once_with(
            "Unable to generate URL from <b>Item Name</b> selections: missing host"
        )

    def test_missing_port_is_caught(self):
        url = {
            "dialect": "mysql",
            "host": "example.com",
            "port": None,
            "database": "trade_nuc",
            "username": "superman",
            "password": "s3cr?t",
        }
        logger = mock.MagicMock()
        self.assertIsNone(convert_to_sqlalchemy_url(url, "Item Name", logger))
        logger.msg_error.emit.assert_called_once_with(
            "Unable to generate URL from <b>Item Name</b> selections: missing port"
        )

    def test_missing_username_is_caught(self):
        url = {
            "dialect": "mysql",
            "host": "example.com",
            "port": 5432,
            "database": "trade_nuc",
            "username": None,
            "password": "s3cr?t",
        }
        logger = mock.MagicMock()
        self.assertIsNone(convert_to_sqlalchemy_url(url, "Item Name", logger))
        logger.msg_error.emit.assert_called_once_with(
            "Unable to generate URL from <b>Item Name</b> selections: missing username"
        )

    def test_missing_password_is_caught(self):
        url = {
            "dialect": "mysql",
            "host": "example.com",
            "port": 5432,
            "database": "trade_nuc",
            "username": "superman",
            "password": None,
        }
        logger = mock.MagicMock()
        self.assertIsNone(convert_to_sqlalchemy_url(url, "Item Name", logger))
        logger.msg_error.emit.assert_called_once_with(
            "Unable to generate URL from <b>Item Name</b> selections: missing password"
        )


class TestConvertUrlToSafeString(unittest.TestCase):
    def test_removes_username_and_password(self):
        url = {
            "dialect": "mysql",
            "host": "example.com",
            "port": 5432,
            "database": "trade_nuc",
            "username": "superman",
            "password": "s3cr?t",
        }
        self.assertEqual(convert_url_to_safe_string(url), "mysql+pymysql://example.com:5432/trade_nuc")

    def test_works_without_credentials_in_url(self):
        url = {
            "dialect": "mysql",
            "host": "example.com",
            "port": 5432,
            "database": "trade_nuc",
        }
        self.assertEqual(convert_url_to_safe_string(url), "mysql+pymysql://example.com:5432/trade_nuc")

    def test_works_with_sqlite_url(self):
        database_path = r"c:\files\database.sqlite" if sys.platform == "win32" else "/dir/data/database.sqlite"
        url = {
            "dialect": "sqlite",
            "database": database_path,
        }
        self.assertEqual(convert_url_to_safe_string(url), r"sqlite:///" + database_path)


class TestSplitUrlCredentials(unittest.TestCase):
    def test_splitting_works(self):
        url = "pymysql://superhuman:verysecret@big.organization.org/d_base"
        sanitized_url, (username, password) = split_url_credentials(url)
        self.assertEqual(sanitized_url, "pymysql://big.organization.org/d_base")
        self.assertEqual(username, "superhuman")
        self.assertEqual(password, "verysecret")


class TestUnsplitUrlCredentials(unittest.TestCase):
    def test_splitting_works(self):
        sanitized_url = "pymysql://big.organization.org/d_base"
        username = "superhuman"
        password = "veryverysecret"
        url = unsplit_url_credentials(sanitized_url, (username, password))
        self.assertEqual(url, "pymysql://superhuman:veryverysecret@big.organization.org/d_base")


if __name__ == "__main__":
    unittest.main()
